from ..enums import Item
from ..utils import Vector, Yaw
from ..buildings import ConveyorBeltMKI, ConveyorBeltMKII, ConveyorBeltMKIII
from ..buildings import Sorter, SorterMKI
from ..buildings import ArcSmelter, PlaneSmelter, NegentrophySmelter
from ..buildings import AssemblingMachineMKI, AssemblingMachineMKII, AssemblingMachineMKIII, ReComposingAssembler
from ..buildings import MatrixLab, SelfEvolutionLab
from ..buildings import OilRefinary
from ..buildings import ChemicalPlant, QuantumChemicalPlant

class FactoryBlock:

    def __init__(self, pos, belt_routing, factory_type, width, recipe):

        factory_pos = pos + FactoryBlock.get_factory_offset(factory_type)
        self.generate_factory(factory_pos, factory_type, width, recipe)

        top_belt_pos = pos + FactoryBlock.get_top_belt_offset(factory_type)
        buttom_belt_pos = pos + FactoryBlock.get_buttom_belt_offset(factory_type)
        self.generate_belts(belt_routing, top_belt_pos, buttom_belt_pos, self.factory, width)

    def generate_belts(self, belt_routing, top_belt_pos, buttom_belt_pos, factory, width):
        top_sorter_count = 0
        buttom_sorter_count = 0
        
        self.ingredient_belts = []
        self.product_belts = []
        
        for route in belt_routing:
            if route.placement == "top":
                pos = top_belt_pos + Vector(y = route.belt_index)
            else:
                pos = buttom_belt_pos - Vector(y = route.belt_index)
            
            if route.direction == "product":
                yaw = Yaw.West
                pos += Vector(x = width - 1)
            elif route.direction == "ingredient":
                yaw = Yaw.East
            else:
                assert False, f"Unknown direction: {route.direction}"
            
            belts = ConveyorBeltMKI.generate_belt(
                name = "",
                pos = pos,
                yaw = yaw,
                length = width
            )
            
            # TODO: Determine the optimal sorter type
            sorter_type = SorterMKI
            
            if route.direction == "product":
                sorter_belt_index = width - 1 - route.belt_index
                sorter_type.generate_sorter_from_belt_to_building(
                    name = "Sorter",
                    belt = belts[sorter_belt_index],
                    building = factory
                )
                self.product_belts.append(belts)
            elif route.direction == "ingredient":
                sorter_belt_index = route.belt_index
                sorter_type.generate_sorter_from_building_to_belt(
                    name = "Sorter",
                    building = factory,
                    belt = belts[sorter_belt_index]
                )
                self.ingredient_belts.append(belts)
            else:
                assert False, f"Unknown direction: {route.direction}"

    def generate_factory(self, pos, factory_type, width, recipe):
        self.factory = factory_type(
            name = "FactoryBlock",
            pos = pos
        )
        
    def connect_to_factory_block(factory_block1, factory_block2):
        for i in range(len(factory_block1.ingredient_belts)):
            factory_block1.ingredient_belts[i][-1].connect_to_belt(factory_block2.ingredient_belts[i][0])
        for i in range(len(factory_block1.product_belts)):
            factory_block2.product_belts[i][-1].connect_to_belt(factory_block1.product_belts[i][0])
    
    def get_inserter_offset(factory_type, side, index):
        assert side == "top" or side == "buttom", "Side needs to be \"top\" or \"buttom\""
        if factory_type == Item.ArcSmelter or factory_type == Item.PlaneSmelter or factory_type == Item.NegentrophySmelter:
            return Vector(
                x = -0.8 + 0.8 * index,
                y = 1.2 if side == "top" else -1.2
            )
        elif factory_type == Item.AssemblingMachineMKI or factory_type == Item.AssemblingMachineMKII or factory_type == Item.AssemblingMachineMKIII or factory_type == Item.ReComposingAssembler:
            return Vector(
                x = -0.8 + 0.8 * index,
                y = 1.2 if side == "buttom" else -1.2    
            )
        elif factory_type == Item.MatrixLab or factory_type == Item.SelfEvolutionLab:
            assert True, "Matrix labs isn't supported yet"
        elif factory_type == Item.OilRefinary:
            assert True, "Oil refinaries isn't supported yet"
        elif factory_type == Item.ChemicalPlant or factory_type == Item.QuantumChemicalPlant:
            assert True, "Chemical plants labs isn't supported yet"
        else:
            assert True, "Unsupported factory type: " + factory_type
    
    def get_inserter_slot(factory_type, side, index):
        assert side == "top" or side == "buttom" or side == "left" or side == "right", "Side needs to be \"top\", \"buttom\", \"left\" or \"right\""
        if factory_type == Item.ArcSmelter or factory_type == Item.PlaneSmelter or factory_type == Item.NegentrophySmelter:
            if side == "top":
                return index
            else:
                return 8 - index
        elif factory_type == Item.AssemblingMachineMKI or factory_type == Item.AssemblingMachineMKII or factory_type == Item.AssemblingMachineMKIII or factory_type == Item.ReComposingAssembler:
            if side == "buttom":
                return index
            else:
                return 8 - index
        elif factory_type == Item.MatrixLab or factory_type == Item.SelfEvolutionLab:
            assert True, "Matrix labs isn't supported yet"
        elif factory_type == Item.OilRefinary:
            assert True, "Oil refinaries isn't supported yet"
        elif factory_type == Item.ChemicalPlant or factory_type == Item.QuantumChemicalPlant:
            assert True, "Chemical plants labs isn't supported yet"
        else:
            assert True, "Unsupported factory type: " + factory_type
    
    def get_belt_index_offset(factory_type):
        if factory_type == Item.ArcSmelter or factory_type == Item.PlaneSmelter or factory_type == Item.NegentrophySmelter:
            return 0
        elif factory_type == Item.AssemblingMachineMKI or factory_type == Item.AssemblingMachineMKII or factory_type == Item.AssemblingMachineMKIII or factory_type == Item.ReComposingAssembler:
            return 0
        elif factory_type == Item.MatrixLab or factory_type == Item.SelfEvolutionLab:
            assert True, "Matrix labs isn't supported yet"
        elif factory_type == Item.OilRefinary:
            assert True, "Oil refinaries isn't supported yet"
        elif factory_type == Item.ChemicalPlant or factory_type == Item.QuantumChemicalPlant:
            assert True, "Chemical plants labs isn't supported yet"
        else:
            assert True, "Unsupported factory type: " + factory_type
    
    def get_belt_offset(factory_type, side):
        if side == "top":
            return Vector(y = self.get_top_belt_y_offset(factory_type))
        elif side == "buttom":
            return Vector(y = self.get_buttom_belt_y_offset(factory_type))
    
    def get_top_belt_offset(factory_type):
        if factory_type == ArcSmelter or factory_type == PlaneSmelter or factory_type == NegentrophySmelter:
            return Vector(y = 2)
        elif factory_type == AssemblingMachineMKI or factory_type == AssemblingMachineMKII or factory_type == AssemblingMachineMKIII or factory_type == ReComposingAssembler:
            return Vector(y = 2)
        elif factory_type == MatrixLab or factory_type == SelfEvolutionLab:
            assert True, "Matrix labs isn't supported yet"
        elif factory_type == OilRefinary:
            assert True, "Oil refinaries isn't supported yet"
        elif factory_type == ChemicalPlant or factory_type == QuantumChemicalPlant:
            assert True, "Chemical plants labs isn't supported yet"
        else:
            assert True, "Unsupported factory type: " + factory_type
    
    def get_buttom_belt_offset(factory_type):
        if factory_type == ArcSmelter or factory_type == PlaneSmelter or factory_type == NegentrophySmelter:
            return Vector(y = -2)
        elif factory_type == AssemblingMachineMKI or factory_type == AssemblingMachineMKII or factory_type == AssemblingMachineMKIII or factory_type == ReComposingAssembler:
            return Vector(y = -2)
        elif factory_type == MatrixLab or factory_type == SelfEvolutionLab:
            assert True, "Matrix labs isn't supported yet"
        elif factory_type == OilRefinary:
            assert True, "Oil refinaries isn't supported yet"
        elif factory_type == ChemicalPlant or factory_type == QuantumChemicalPlant:
            assert True, "Chemical plants labs isn't supported yet"
        else:
            assert True, "Unsupported factory type: " + factory_type
        
    def get_factory_offset(factory_type):
        if factory_type == ArcSmelter or factory_type == PlaneSmelter or factory_type == NegentrophySmelter:
            return Vector(x = 1)
        elif factory_type == AssemblingMachineMKI or factory_type == AssemblingMachineMKII or factory_type == AssemblingMachineMKIII or factory_type == ReComposingAssembler:
            return Vector(x = 1)
        elif factory_type == MatrixLab or factory_type == SelfEvolutionLab:
            assert True, "Matrix labs isn't supported yet"
        elif factory_type == OilRefinary:
            assert True, "Oil refinaries isn't supported yet"
        elif factory_type == ChemicalPlant or factory_type == QuantumChemicalPlant:
            assert True, "Chemical plants labs isn't supported yet"
        else:
            assert True, "Unsupported factory type: " + factory_type
    


from .building import Building, Factory3x3
from ..utils import Vector, Yaw
from ..enums import BuildingItem, BuildingModel

class AssemblingMachine(Factory3x3):
    def __init__(self, name, pos: Vector, item_id: BuildingItem, model_index: int, recipe_id: int = 0):
        super().__init__(name)
        self.pos = pos
        self.pos2 = pos
        self.yaw = Yaw.North
        self.yaw2 = Yaw.North
        self.item_id = item_id
        self.model_index = model_index
        self.output_object_index = -1
        self.input_object_index = -1
        self.recipe_id = recipe_id
        self.parameter_count = 1
        self.parameters = [0]
        
    def get_size():
        return Vector(4.0, 4.0)
    
    def get_offset():
        return Vector(2.0, 2.0)
        
class AssemblingMachineMKI(AssemblingMachine):
    def __init__(self, name, pos: Vector, recipe_id: int = 0):
        super().__init__(
            name = name,
            pos = pos,
            item_id = BuildingItem.AssemblingMachineMKI,
            model_index = BuildingModel.AssemblingMachineMKI,
            recipe_id = recipe_id
        )
            
class AssemblingMachineMKII(AssemblingMachine):
    def __init__(self, name, pos: Vector, recipe_id: int = 0):
        super().__init__(
            name = name,
            pos = pos,
            item_id = BuildingItem.AssemblingMachineMKII,
            model_index = BuildingModel.AssemblingMachineMKII,
            recipe_id = recipe_id
        )
            
class AssemblingMachineMKIII(AssemblingMachine):
    def __init__(self, name, pos: Vector, recipe_id: int = 0):
        super().__init__(
            name = name,
            pos = pos,
            item_id = BuildingItem.AssemblingMachineMKIII,
            model_index = BuildingModel.AssemblingMachineMKIII,
            recipe_id = recipe_id
        )
            
class ReComposingAssembler(AssemblingMachine):
    def __init__(self, name, pos: Vector, recipe_id: int = 0):
        super().__init__(
            name = name,
            pos = pos,
            item_id = BuildingItem.ReComposingAssembler,
            model_index = BuildingModel.ReComposingAssembler,
            recipe_id = recipe_id
        )


from .packet import Packet
from ..utils import Vector

class BlueprintHeader:

    def __init__(self, size: Vector = Vector()):
        self.version = 1
        self.cursor_offset_x = int(size.x // 2)
        self.cursor_offset_y = int(size.y // 2)
        self.cursor_target_area = 0
        self.dragbox_size_x = int(size.x)
        self.dragbox_size_y = int(size.y)
        self.primary_area_index = 0
        self.area_count = 1

    def parse(self, packet):
        self.version = packet.parse_int32()
        self.cursor_offset_x = packet.parse_int32()
        self.cursor_offset_y = packet.parse_int32()
        self.cursor_target_area = packet.parse_int32()
        self.dragbox_size_x = packet.parse_int32()
        self.dragbox_size_y = packet.parse_int32()
        self.primary_area_index = packet.parse_int32()
        self.area_count = packet.parse_int8()

    def serialize(self):
        packet = Packet()
        packet.serialize_int32(self.version)
        packet.serialize_int32(self.cursor_offset_x)
        packet.serialize_int32(self.cursor_offset_y)
        packet.serialize_int32(self.cursor_target_area)
        packet.serialize_int32(self.dragbox_size_x)
        packet.serialize_int32(self.dragbox_size_y)
        packet.serialize_int32(self.primary_area_index)
        packet.serialize_int8(self.area_count)
        return packet

    def __str__(self):
        return f"""
Blueprint header:
=================
Binary data: {self.serialize().data}
=================
Version: {self.version}
Cursor offset x: {self.cursor_offset_x}
Cursor offset y: {self.cursor_offset_y}
Cursor target area: {self.cursor_target_area}
Dragbox size x: {self.dragbox_size_x}
Dragbox size y: {self.dragbox_size_y}
Primary area index: {self.primary_area_index}
Area count: {self.area_count}
"""


from .packet import Packet

class BlueprintBuildingHeader:

    def __init__(self, building_count = 0):
        self.building_count = building_count

    def parse(self, packet):
        self.building_count = packet.parse_int32()

    def serialize(self):
        packet = Packet()
        packet.serialize_int32(self.building_count)
        return packet

    def __str__(self):
        return f"""
Blueprint Building Header:
==========================
Binary data: {self.serialize().data}
==========================
Count: {self.building_count}
"""


from .building import Building
from ..utils import Vector, Yaw
from ..enums import BuildingItem, BuildingModel

class Sorter(Building):
    
    def __init__(self, name, pos: Vector, pos2: Vector, yaw: Yaw, output_object_index: int = -1, input_object_index: int = -1, output_to_slot: int = -1, input_from_slot: int = -1, output_from_slot: int = -1, input_to_slot: int = -1, output_offset: int = -1, input_offset: int = -1, parameters = []):
        super().__init__(name)
        self.pos = pos
        self.pos2 = pos2
        self.yaw = yaw
        self.yaw2 = yaw
        self.item_id = -1
        self.model_index = -1
        self.output_object_index = output_object_index
        self.input_object_index = input_object_index
        self.output_to_slot = output_to_slot
        self.input_from_slot = input_from_slot
        self.output_from_slot = output_from_slot
        self.input_to_slot = input_to_slot
        self.output_offset = output_offset
        self.input_offset = input_offset
        self.parameter_count = len(parameters)
        self.parameters = parameters

    def set_filter(self, item):
        assert False, "Not supported"

    def generate_sorter_from_belt_to_building(name, belt, building, sorter_type):
        yaw = Yaw.get_neares_90_degree(belt.pos, building.pos)
        slot = building.get_nearest_slot_from_position(belt.pos)
        slot_pos = building.get_position_of_slot(slot)
        return sorter_type(
            name = name,
            pos = belt.pos,
            pos2 = slot_pos,
            yaw = yaw,
            output_object_index = building.index,
            input_object_index = belt.index,
            output_to_slot = slot,
            input_from_slot = -1,
            output_from_slot = 0,
            input_to_slot = 1,
            output_offset = 0,
            input_offset = 0,
            parameters = [1]
        )

    def generate_sorter_from_building_to_belt(name, building, belt, sorter_type = None):
        yaw = Yaw.get_neares_90_degree(building.pos, belt.pos)
        slot = building.get_nearest_slot_from_position(belt.pos)
        slot_pos = building.get_position_of_slot(slot)
        return sorter_type(
            name = name,
            pos = slot_pos,
            pos2 = belt.pos,
            yaw = yaw,
            output_object_index = belt.index,
            input_object_index = building.index,
            output_to_slot = -1,
            input_from_slot = slot,
            output_from_slot = 0,
            input_to_slot = 1,
            output_offset = -1,
            input_offset = 0,
            parameters = [1]
        )

    def generate_sorter_from_building_to_building(name, building_1, building_2, sorter_type = None):
        # Only one sorter can be generated from one factory to another
        yaw = Yaw.get_neares_90_degree(building_1.pos, building_2.pos)
        slot_1 = building_1.get_nearest_slot_from_position(building_2.pos)
        slot_2 = building_2.get_nearest_slot_from_position(building_1.pos)
        slot_1_pos = building_1.get_position_of_slot(slot_1)
        slot_2_pos = building_2.get_position_of_slot(slot_2)
        return sorter_type(
            name = name,
            pos = slot_1_pos,
            pos2 = slot_2_pos,
            yaw = yaw,
            output_object_index = building_2.index,
            input_object_index = building_1.index,
            output_to_slot = slot_2,
            input_from_slot = slot_1,
            output_from_slot = 0,
            input_to_slot = 1,
            output_offset = 0,
            input_offset = 0,
            parameters = [1]
        )

    def generate_sorter_from_belt_to_belt(name, factory, belt, sorter_type = None):
        assert False, "Not supported"

class SorterMKI(Sorter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.item_id = BuildingItem.SorterMKI
        self.model_index = BuildingModel.SorterMKI
        
    def generate_sorter_from_belt_to_building(name, belt, building):
        return Sorter.generate_sorter_from_belt_to_building(name, belt, building, SorterMKI)
    
    def generate_sorter_from_building_to_belt(name, building, belt):
        return Sorter.generate_sorter_from_building_to_belt(name, building, belt, SorterMKI)
    
    def generate_sorter_from_building_to_building(name, building_1, building_2):
        return Sorter.generate_sorter_from_building_to_building(name, building_1, building_2, SorterMKI)

class SorterMKII(Sorter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.item_id = BuildingItem.SorterMKII
        self.model_index = BuildingModel.SorterMKII
                
    def generate_sorter_from_belt_to_building(name, belt, building):
        return Sorter.generate_sorter_from_belt_to_building(name, belt, building, SorterMKII)
    
    def generate_sorter_from_building_to_belt(name, building, belt):
        return Sorter.generate_sorter_from_building_to_belt(name, building, belt, SorterMKII)

    def generate_sorter_from_building_to_building(name, building_1, building_2):
        return Sorter.generate_sorter_from_building_to_building(name, building_1, building_2, SorterMKII)

class SorterMKIII(Sorter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.item_id = BuildingItem.SorterMKIII
        self.model_index = BuildingModel.SorterMKIII
                
    def generate_sorter_from_belt_to_building(name, belt, building):
        return Sorter.generate_sorter_from_belt_to_building(name, belt, building, SorterMKIII)
    
    def generate_sorter_from_building_to_belt(name, building, belt):
        return Sorter.generate_sorter_from_building_to_belt(name, building, belt, SorterMKIII)

    def generate_sorter_from_building_to_building(name, building_1, building_2):
        return Sorter.generate_sorter_from_building_to_building(name, building_1, building_2, SorterMKIII)

class PileSorter(Sorter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.item_id = BuildingItem.PileSorter
        self.model_index = BuildingModel.PileSorter
                
    def generate_sorter_from_belt_to_building(name, belt, building):
        return Sorter.generate_sorter_from_belt_to_building(name, belt, building, PileSorter)
    
    def generate_sorter_from_building_to_belt(name, building, belt):
        return Sorter.generate_sorter_from_building_to_belt(name, building, belt, PileSorter)

    def generate_sorter_from_building_to_building(name, building_1, building_2):
        return Sorter.generate_sorter_from_building_to_building(name, building_1, building_2, PileSorter)



from .factory_line import FactoryLine
from .belt_router import BeltRouter
from ..utils import Vector
import enum

class FactorySection:
    
    class BeltRouting:
        
        def __init__(self, name, count_per_second, direction, placement, router_index, belt_index):
            self.name = name
            self.count_per_second = count_per_second
            self.direction = direction
            self.placement = placement
            self.router_index = router_index
            self.belt_index = belt_index
    
    def __init__(self, pos, input_count, output_count, main_belts, product, recipe):
        self.product_count = len(recipe["output_items"].keys())
        
        self.router_width = 2 * (input_count + output_count + self.product_count)
        self.height = 7

        belt_routing = FactorySection.get_belt_routing(product, main_belts, recipe)
        
        # Create factory line
        factory_line_pos = pos + Vector(x = self.router_width)
        factory_count = 5
        self.factory_line = FactoryLine(factory_line_pos, belt_routing, recipe, factory_count)
        
        # Create router
        self.router = BeltRouter(pos, input_count, output_count, self.product_count, belt_routing, self.height)
        
        # Connect factory line and router
        self.connect_to_factory_line(len(recipe["input_items"].keys()), self.product_count)
        
    def get_belt_routing(product, main_belts, recipe):
        ingredients = product.get_needed_ingredients()
        normalized_products = recipe["output_items"]
        products = product.get_products()
        
        routes = []
        for ingredient in ingredients:
            routes.append(
                FactorySection.BeltRouting(
                    name = ingredient.name,
                    count_per_second = ingredient.count_pr_sec,
                    direction = "ingredient",
                    placement = "top",
                    router_index = 0,
                    belt_index = 0
                )
            )
            
        for product in products:
            routes.append(
                FactorySection.BeltRouting(
                    name = product.name,
                    count_per_second = product.count_pr_sec,
                    direction = "product",
                    placement = "top",
                    router_index = 0,
                    belt_index = 0
                )
            )
        
        routes.sort(reverse = True, key = lambda flow: flow.count_per_second)
        
        top_count = 0
        buttom_count = 0
        for i, route in enumerate(routes):
            if i % 2 == 0:
                route.placement = "top"
                route.belt_index = top_count
                top_count += 1
            else:
                route.placement = "buttom"
                route.belt_index = buttom_count
                buttom_count += 1
            for i, belt in enumerate(main_belts):
                if belt.name == route.name:
                    route.router_index = i
                    break
            
        return routes
    
    def connect_to_factory_line(self, factory_input_count, factory_output_count):
        #for i in range(factory_input_count):
        #    self.router.selector_belts[i][-1].connect_to_belt(self.factory_line.factory_blocks[0].ingredient_belts[factory_input_count - i - 1][0])
        
        """
        for i in range(factory_output_count):
            print()
            self.factory_line.factory_blocks[0].product_belts[i][-1].connect_to_belt(self.router.product_belts[i][0])
        """    
    def connect_to_section(self, section2):
        assert \
            len(self.router.input_belts) == len(section2.router.input_splitters) and \
            len(self.router.output_belts) == len(section2.router.output_splitters), \
                "The two sections must have same dimensions of input- and output belts"
        
        for i in range(len(self.router.input_belts)):
            section2.router.input_splitters[i].connect_to_belt(self.router.input_belts[i])
        for i in range(len(self.router.output_belts)):
            self.router.output_belts[i].connect_to_splitter(section2.router.output_splitters[i])


from .building import Building
from ..utils import Yaw
from ..enums import BuildingItem, BuildingModel

class PowerDistributor(Building):
    def __init__(self, name, pos):
        super().__init__(name)
        self.pos = pos
        self.pos2 = pos
        self.yaw = Yaw.North
        self.yaw2 = Yaw.North
        self.item_id = None
        self.model_index = None
        self.output_object_index = -1
        self.input_object_index = -1
        self.output_to_slot = 0
        self.input_from_slot = 0
        self.output_from_slot = 0
        self.input_to_slot = 0
        self.output_offset = 0
        self.input_offset = 0
        self.parameter_count = 0
        self.parameters = []

class TeslaTower(PowerDistributor):
    def __init__(self, name, pos):
        super().__init__(name, pos)
        self.item_id = BuildingItem.TeslaTower
        self.model_index = BuildingModel.TeslaTower

    def get_size(self):
        return Vector(1.0, 1.0)
    
    def get_offset(self):
        return Vector(0.5, 0.5)

class WirelessPowerTower(PowerDistributor):
    def __init__(self, name, pos):
        super().__init__(name, pos)
        self.item_id = BuildingItem.WirelessPowerTower
        self.model_index = BuildingModel.WirelessPowerTower
        
    def get_size(self):
        return Vector(1.0, 1.0)
    
    def get_offset(self):
        return Vector(0.5, 0.5)

class SateliteSubstation(PowerDistributor):
    def __init__(self, name, pos):
        super().__init__(name, pos)
        self.item_id = BuildingItem.SateliteSubstation
        self.model_index = BuildingModel.SateliteSubstation

    def get_size(self):
        return Vector(4.0, 4.0)
    
    def get_offset(self):
        return Vector(2.0, 2.0)



from .blueprint import Blueprint
from .blueprint_string_header import BlueprintStringHeader
from .blueprint_header import BlueprintHeader
from .blueprint_area import BlueprintArea
from .blueprint_building_header import BlueprintBuildingHeader
from .blueprint_building import BlueprintBuilding, BlueprintBuildingV1, BlueprintBuildingV2


from .building import Building, Factory3x3
from .assembling_machine import AssemblingMachineMKI, AssemblingMachineMKII, AssemblingMachineMKIII, ReComposingAssembler
from .conveyer_belt import ConveyorBeltMKI, ConveyorBeltMKII, ConveyorBeltMKIII
from .sorter import Sorter, SorterMKI, SorterMKII, SorterMKIII, PileSorter
from .splitter import Splitter
from .power_distributor import TeslaTower, WirelessPowerTower, SateliteSubstation
from .smelter import ArcSmelter, PlaneSmelter, NegentrophySmelter
from .lab import MatrixLab, SelfEvolutionLab
from .oil_refinary import OilRefinary
from .chemical_plant import ChemicalPlant, QuantumChemicalPlant
from .depot import DepotMKI, DepotMKII, LogisticDistributor


import math

class Vector:

    def __init__(self, x = 0.0, y = 0.0, z = 0.0):
        self.x = x
        self.y = y
        self.z = z

    def move(self, offset):
        self.x += offset.x
        self.y += offset.y
        self.z += offset.z

    def get_distance(self, pos):
        return math.sqrt(math.pow(pos.x - self.x, 2) + math.pow(pos.y - self.y, 2))

    def __add__(pos1, pos2):
        return Vector(
            x = pos1.x + pos2.x,
            y = pos1.y + pos2.y,
            z = pos1.z + pos2.z
        )

    def __sub__(pos1, pos2):
        return Vector(
            x = pos1.x - pos2.x,
            y = pos1.y - pos2.y,
            z = pos1.z - pos2.z
        )
        
    def __str__(self):
        return f"x: {self.x}\ny: {self.y}\nz: {self.z}\n"

class Yaw:
    North = 0.0
    NorthEast = 45.0
    East = 90.0
    SouthEast = 135.0
    South = 180.0
    SouthWest = 225.0
    West = 270.0
    NorthWest = 315.0
    Unknown = None

    def get_neares_90_degree(pos1: Vector, pos2: Vector):
        angle = Yaw.get_angle(pos1, pos2)
        if angle == None:
            return None
        angle = (angle + 45) // 90 * 90
        if angle == 360.00:
            angle = 0
        return angle

    def get_angle(pos1, pos2):
        delta_x = pos2.x - pos1.x
        delta_y = pos2.y - pos1.y
        
        if delta_x == 0:
            if delta_y > 0:
                return Yaw.North
            elif delta_y < 0:
                return Yaw.South
            else:
                return Yaw.Unknown
        angle = math.degrees(math.atan2(delta_x, delta_y))
        
        if angle < 0.0:
            angle += 360.0
        return angle
            
    def direction_average(dir1, dir2):
        if (dir1 == Yaw.North and dir2 == Yaw.East) or (dir1 == Yaw.East and dir2 == Yaw.North):
            return Yaw.NorthEast
        elif (dir1 == Yaw.East and dir2 == Yaw.South) or (dir1 == Yaw.South and dir2 == Yaw.East):
            return Yaw.SouthEast
        elif (dir1 == Yaw.South and dir2 == Yaw.West) or (dir1 == Yaw.West and dir2 == Yaw.South):
            return Yaw.SouthWest
        elif (dir1 == Yaw.West and dir2 == Yaw.North) or (dir1 == Yaw.North and dir2 == Yaw.West):
            return Yaw.NorthWest
        else:
            return None

    def direction_to_unit_vector(direction):
        if direction == Yaw.North:
            return 0.0, 1.0
        elif direction == Yaw.East:
            return 1.0, 0.0
        elif direction == Yaw.South:
            return 0.0, -1.0
        elif direction == Yaw.West:
            return -1.0, 0.0
        return None

if __name__ == "__main__":
    string = ""
    for i in range(11):
        for j in range(11):
            angle = Yaw.get_angle(Vector(5, 5), Vector(j, 10 - i))
            if angle != None:
                string += "{:.2f}\t".format(angle)
            else:
                string += "{:.2f}\t".format(0)
        string += "\n"
    print(string)
    string = ""
    for i in range(11):
        for j in range(11):
            angle = Yaw.get_neares_90_degree(Vector(5, 5), Vector(j, 10 - i))
            if angle != None:
                string += "{:.2f}\t".format(angle)
            else:
                string += "{:.2f}\t".format(0)
        string += "\n"
    print(string)


from struct import pack, unpack
from copy import copy

class Packet:

    def __init__(self, data = b""):
        self.data = copy(data)

    def parse_int8(self):
        value = unpack("b", self.data[0:1])[0]
        self.data = self.data[1:]
        return value

    def parse_int16(self):
        value = unpack("h", self.data[0:2])[0]
        self.data = self.data[2:]
        return value

    def parse_int32(self):
        value = unpack("i", self.data[0:4])[0]
        self.data = self.data[4:]
        return value

    def parse_float32(self):
        value = unpack("f", self.data[0:4])[0]
        self.data = self.data[4:]
        return value

    def peak_parse_int8(self):
        value = unpack("b", self.data[0:1])[0]
        return value

    def peak_parse_int16(self):
        value = unpack("h", self.data[0:2])[0]
        return value

    def peak_parse_int32(self):
        value = unpack("i", self.data[0:4])[0]
        return value

    def peak_parse_float32(self):
        value = unpack("f", self.data[0:4])[0]
        return value

    def serialize_int8(self, value):
        self.data += pack("b", value)

    def serialize_int16(self, value: int):
        self.data += pack("h", value)

    def serialize_int32(self, value: int):
        self.data += pack("i", value)

    def serialize_float32(self, value: float):
        self.data += pack("f", value)



from .building import Factory
from ..utils import Vector, Yaw
from ..enums import BuildingItem, BuildingModel

class OilRefinary(Factory):

#         ┌───────────────────────┐
#         │                       │
#         │                       │
#         │                       │
#         │-----------------------│
#         │                       │
# slot 2  │                       │ slot 3
#         │                       │
#         │-----------------------│
#         │                       │
# slot 1  │           X           │ slot 4
#         │                       │
#         │-----------------------│
#         │                       │
# slot 0  │                       │ slot 5
#         │                       │
#         │-----------------------│
#         │                       │
#         │                       │
#         │                       │
#         │-----------------------│
#         │                       │
#         │                       │
#         │                       │
#         └───────────────────────┘
#           slot 8  slot 7  slot 6  

    def __init__(self, name, pos: Vector, recipe_id: int = 0):
        super().__init__(name)
        self.pos = pos
        self.pos2 = pos
        self.yaw = Yaw.North
        self.yaw2 = Yaw.North
        self.item_id = BuildingItem.OilRefinary
        self.model_index = BuildingModel.OilRefinary
        self.output_object_index = -1
        self.input_object_index = -1
        self.recipe_id = recipe_id
        self.parameter_count = 1
        self.parameters = [0]

    def number_of_slots(self):
        return 9

    def get_position_of_slot(self, slot):
        assert slot >= 0 and slot < self.number_of_slots(), f"slot index needs to be: slot >= 0 and slot < {self.number_of_slots()} (slot was {slot})"
        if slot == 0:
            delta_pos = Vector(x = 0.877, y = -0.8)
        elif slot == 1:
            delta_pos = Vector(x = 0.877, y = 0.0)
        elif slot == 2:
            delta_pos = Vector(x = 0.877, y = 0.8)
        elif slot == 3:
            delta_pos = Vector(x = -0.877, y = 0.8)
        elif slot == 4:
            delta_pos = Vector(x = -0.877, y = 0.0)
        elif slot == 5:
            delta_pos = Vector(x = -0.877, y = -0.8)
        elif slot == 6:
            delta_pos = Vector(x = -0.9, y = -2.862)
        elif slot == 7:
            delta_pos = Vector(x = 0.0, y = -2.862)
        elif slot == 8:
            delta_pos = Vector(x = 0.9, y = -2.862)
        return delta_pos + self.pos



#import Buildings
from ..utils import Vector, Yaw
from ..buildings import ConveyorBeltMKI, Splitter
from ..enums import BuildingModel

class BeltRouter:

    def __init__(self, pos, input_count, output_count, product_count, belt_routing, belt_length):
        self.width = 2 * (input_count + output_count + product_count)
        
        self.generate_splitters(pos, input_count, output_count, product_count)
        self.generate_input_belts(pos, input_count, belt_length)
        self.generate_output_belts(pos, input_count, output_count, product_count, belt_length)
        self.generate_router_belts(pos, belt_routing)
        
    def generate_splitters(self, pos, input_count, output_count, product_count):
        self.input_splitters = []
        for i in range(input_count):
            temp_pos = Vector(pos.x + 2 * i, pos.y)
            input_splitter = Splitter(
                name = f"InputSplitter:{i}",
                pos = temp_pos,
                yaw = Yaw.North,
                mode = BuildingModel.SplitterTwoLayerStraight
            )
            self.input_splitters.append(input_splitter)
        self.output_splitters = []
        for i in range(output_count):
            temp_pos = Vector(pos.x + 2 * (i + input_count), pos.y)
            output_splitter = Splitter(
                name = f"OutputSplitter:{i}",
                pos = temp_pos,
                yaw = Yaw.North,
                mode = BuildingModel.SplitterTwoLayerStraight
            )
            self.output_splitters.append(output_splitter)
        self.product_splitters = []
        for i in range(product_count):
            temp_pos = Vector(pos.x + 2 * (i + input_count + output_count), pos.y)
            product_splitter = Splitter(
                name = f"ProductSplitter:{i}",
                pos = temp_pos,
                yaw = Yaw.North,
                mode = BuildingModel.SplitterTwoLayerStraight
            )
            self.product_splitters.append(product_splitter)
        self.splitters = self.input_splitters + self.output_splitters + self.product_splitters
        
    def generate_input_belts(self, pos, input_count, belt_length):
        self.input_belts = []
        for i in range(input_count):
            temp_pos = Vector(pos.x + 2 * i, pos.y + belt_length - 1, 1)
            belts = ConveyorBeltMKI.generate_belt(f"BeltRouter:InputBelt:{i}", temp_pos, Yaw.South, belt_length)
            belt_start = belts[0]
            belt_end = belts[-1]
            self.input_belts.append(belt_start)
            ### Connect belts to splitters
            splitter = self.input_splitters[i]
            belt_end.connect_to_splitter(splitter)
            
    def generate_output_belts(self, pos, input_count, output_count, product_count, belt_length):
        self.output_belts = []
        for i in range(output_count):
            temp_pos = Vector(pos.x + 2 * (i + input_count), pos.y, 1)
            belts = ConveyorBeltMKI.generate_belt(f"BeltRouter:OutputBelt:{i}", temp_pos, Yaw.North, belt_length)
            belt_start = belts[0]
            belt_end = belts[-1]
            ## Connect belts to splitters
            self.output_belts.append(belt_end)
            splitter = self.output_splitters[i]
            splitter.connect_to_belt(belt_start)
        for i in range(product_count):
            temp_pos = Vector(pos.x + 2 * (i + input_count + output_count), pos.y, 1)
            belts = ConveyorBeltMKI.generate_belt(f"BeltRouter:ProductBelt:{i}", temp_pos, Yaw.North, belt_length)
            belt_start = belts[0]
            belt_end = belts[-1]
            ## Connect belts to splitters
            self.output_belts.append(belt_end)
            splitter = self.product_splitters[i]
            splitter.connect_to_belt(belt_start)
    
    def generate_router_belts(self, pos, routes):
        self.product_belts = []
        self.selector_belts = []
        for route in routes:
            if route.direction == "product":
                if route.placement == "top":
                    start_pos = pos + Vector(self.width, 2 + route.belt_index)
                    yaw = [Yaw.West, Yaw.South]
                else:
                    start_pos = pos + Vector(self.width - 1, -2 - route.belt_index)
                    yaw = [Yaw.West, Yaw.North]
                name = "Productbelt"
                length = [self.width - 1 - 2 * route.router_index, 3 + route.belt_index]
                self.product_belts.append(ConveyorBeltMKI.generate_belt(name, start_pos, yaw, length))
                splitter = self.splitters[route.router_index]
                self.product_belts[-1][-1].connect_to_splitter(splitter)
                
                for i in range(route.belt_index):
                    self.product_belts[-1][-2-i].move_relative(Vector(z = 0.0))
                
                # Raise belt to allow crossing belts to go under
                for i in range(route.belt_index):
                    self.product_belts[-1][-3 - i].move_relative(Vector(z = 0.3))
                
            elif route.direction == "ingredient":
                if route.placement == "top":
                    start_pos = pos + Vector(route.router_index * 2, 0)
                    yaw = [Yaw.North, Yaw.East]
                else:
                    start_pos = pos + Vector(route.router_index * 2, 0)
                    yaw = [Yaw.South, Yaw.East]
                
                name = "SelectorBelt",
                length = [2 + route.belt_index, self.width - 2 * route.router_index]
                self.selector_belts.append(ConveyorBeltMKI.generate_belt(name, start_pos, yaw, length))
                splitter = self.splitters[route.router_index]
                splitter.connect_to_belt(self.selector_belts[-1][0])
                
                # Raise belt to allow crossing belts to go under
                for i in range(route.belt_index):
                    self.selector_belts[-1][2 + i].move_relative(Vector(z = 0.3))



from .building import Building
from ..utils import Vector, Yaw
from ..enums import BuildingItem, BuildingModel

class ConveyorBelt(Building):
    def __init__(self, name, pos: Vector, yaw: Yaw, output_object_index: int = -1, input_object_index: int = -1, output_to_slot: int = 0):
        super().__init__(name)
        self.pos = pos
        self.pos2 = pos
        self.yaw = yaw
        self.yaw2 = yaw
        self.output_object_index = output_object_index
        self.output_to_slot = output_to_slot
        
        self.output_from_slot = 0
        self.input_to_slot = 1
        self.input_from_slot = 0
        
    def connect_to_belt(belt1, belt2):
        dx, dy = Yaw.direction_to_unit_vector(belt1.yaw)
        if ((int(belt1.pos.x + dx) == belt2.pos.x) and (int(belt1.pos.y + dy) == belt2.pos.y) and belt1.yaw == belt2.yaw):
            belt1.output_object_index = belt2.index
            belt1.output_to_slot = 1
        else:
            dx, dy = Yaw.direction_to_unit_vector(belt2.yaw)
            assert False, "Wrong use of belt-to_belt connection"
            
    def connect_to_splitter(self, splitter):
        assert (int(splitter.yaw - self.yaw) % 180 == 0), f"Wrong orientation (splitter: {splitter.yaw}, belt: {self.yaw})"
        
        self.output_object_index = splitter.index
        self.output_to_slot = 1
        
        if self.yaw == splitter.yaw:
            self.output_to_slot = 2
        else:
            self.output_to_slot = 0
        if self.pos.z == 1:
            self.output_to_slot += 1
        
        # Move the belt back 0.2 spaces
        dx, dy = Yaw.direction_to_unit_vector(self.yaw)
        self.move_relative(Vector(-dx * 0.2, -dy * 0.2))

    def connect_to_sorter(self, sorter):
        pass

    def generate_belt(name, pos, yaw, length: int, belt_type):
        
        if type(yaw) != list:
            yaw = [yaw]
        if type(length) != list:
            length = [length]
        assert len(yaw) == len(length), "\"yaw\" and \"length\" must have the same length"
        
        for l in length:
            assert l > 0, "Belt length is less than one!"
        
        belts = []
        for i in range(len(yaw)):
            dx, dy = Yaw.direction_to_unit_vector(yaw[i])
            for j in range(length[i]):
                belt = belt_type(
                    name = f"{name}:{len(belts)}",
                    pos = pos,
                    yaw = yaw[i],
                    output_object_index = Building.count() + 1,
                    output_to_slot = 1
                )
                belts.append(belt)
                pos += Vector(dx, dy)
            
            if (i != len(yaw) - 1):
                belt.yaw = Yaw.direction_average(yaw[i], yaw[i + 1])
            else:
                belt.output_object_index = -1
                belt.output_to_slot = 0
        return belts
        
class ConveyorBeltMKI(ConveyorBelt):
    def __init__(self, name, pos: Vector, yaw: Yaw, output_object_index: int = -1, input_object_index: int = -1, output_to_slot: int = 0):
        super().__init__(
            name = name,
            pos = pos,
            yaw = yaw,
            output_object_index = output_object_index,
            input_object_index = input_object_index,
            output_to_slot = output_to_slot
        )
        self.item_id = BuildingItem.ConveyorBeltMKI
        self.model_index = BuildingModel.ConveyorBeltMKI
    
    def generate_belt(name, pos, yaw, length: int):
        return ConveyorBelt.generate_belt(name, pos, yaw, length, ConveyorBeltMKI)
        
class ConveyorBeltMKII(ConveyorBelt):
    def __init__(self, name, pos: Vector, yaw: Yaw, output_object_index: int = -1, input_object_index: int = -1, output_to_slot: int = 0):
        super().__init__(
            name = name,
            pos = pos,
            yaw = yaw,
            output_object_index = output_object_index,
            input_object_index = input_object_index,
            output_to_slot = output_to_slot
        )
        self.item_id = BuildingItem.ConveyorBeltMKII
        self.model_index = BuildingModel.ConveyorBeltMKII

    def generate_belt(name, pos, yaw, length: int):
        return ConveyorBelt.generate_belt(name, pos, yaw, length, ConveyorBeltMKII)
        
class ConveyorBeltMKIII(ConveyorBelt):
    def __init__(self, name, pos: Vector, yaw: Yaw, output_object_index: int = -1, input_object_index: int = -1, output_to_slot: int = 0):
        super().__init__(
            name = name,
            pos = pos,
            yaw = yaw,
            output_object_index = output_object_index,
            input_object_index = input_object_index,
            output_to_slot = output_to_slot
        )
        self.item_id = BuildingItem.ConveyorBeltMKIII
        self.model_index = BuildingModel.ConveyorBeltMKIII

    def generate_belt(name, pos, yaw, length: int):
        return ConveyorBelt.generate_belt(name, pos, yaw, length, ConveyorBeltMKIII)
        


from .building import Building
from .conveyer_belt import ConveyorBelt
from ..utils import Vector, Yaw
from ..enums import BuildingItem, BuildingModel

class Splitter(Building):
    
    # Mode should be one of: 
    # - BuildingModel.SplitterCross
    # - BuildingModel.SplitterTwoLayerStraight
    # - BuildingModel.SplitterTwoLayerCross
    def __init__(self, name, pos: Vector, yaw: Yaw, mode):
        super().__init__(name)
        self.pos = pos
        self.pos2 = pos
        self.yaw = yaw
        self.yaw2 = yaw
        self.item_id = BuildingItem.Splitter
        self.model_index = mode
        self.output_object_index = -1
        self.input_object_index = -1
        self.output_to_slot = 14
        self.input_from_slot = 15
        self.output_from_slot = 15
        self.input_to_slot = 14
        self.output_offset = 0
        self.input_offset = 0
        self.parameter_count = 6
        self.parameters = [0, 0, 0, 0, 0, 0]
        
    def connect_to_belt(self, belt: ConveyorBelt):
        assert (int(self.yaw - belt.yaw) % 180 == 0), f"Wrong orientation (splitter: {splitter.yaw}, belt: {belt.yaw})"
        
        belt.input_object_index = self.index
        
        # Set slot index
        if self.yaw == belt.yaw:
            belt.input_from_slot = 0
        else:
            belt.input_from_slot = 2
        if belt.pos.z == 1:
            belt.input_from_slot += 1
        
        dx, dy = Yaw.direction_to_unit_vector(belt.yaw)
        belt.move_relative(Vector(dx * 0.2, dy * 0.2))
    
    def prioritize_input(self, slot):
        assert False, "Not supported"

    def prioritize_output(self, slot):
        assert False, "Not supported"
    
    def set_output_filter(self, slot, item):
        assert False, "Not supported"
        
    def get_size():
        return Vector(2.0, 2.0)
    
    def get_offset():
        return Vector(1.0, 1.0)


from .packet import Packet
from ..utils import Vector
from ..enums import BuildingItem, BuildingModel
from copy import copy
import enum

class BlueprintBuilding:
    
    def __init__(self):
        self.index = 0
        self.area_index = 0
        self.pos = Vector()
        self.pos2 = Vector()
        self.yaw = 0
        self.yaw2 = 0
        self.item_id = 0
        self.model_index = 0
        self.output_object_index = -1
        self.input_object_index = -1
        self.output_to_slot = -1
        self.input_from_slot = -1
        self.output_from_slot = -1
        self.input_to_slot = -1
        self.output_offset = 0
        self.input_offset = 0
        self.recipe_id = 0
        self.filter_id = 0
        self.parameter_count = 0
        self.parameters = []
    
    def move_relative(self, pos: Vector):
        self.pos.x += pos.x
        self.pos.y += pos.y
        self.pos.z += pos.z
        self.pos2.x += pos.x
        self.pos2.y += pos.y
        self.pos2.z += pos.z

    def move_absolute(self, pos: Vector):
        self.pos.x = pos.x
        self.pos.y = pos.y
        self.pos.z = pos.z
        self.pos2.x = pos.x
        self.pos2.y = pos.y
        self.pos2.z = pos.z

    def get_version(packet):
        data = packet.peak_parse_int32()
        if data == -100:
            return BlueprintBuildingV2
        else:
            return BlueprintBuildingV1

class BlueprintBuildingV1(BlueprintBuilding):

    def __init__(self, building):
        super().__init__()
        self.index = building.index
        self.area_index = building.area_index
        self.pos.x = building.pos.x
        self.pos.y = building.pos.y
        self.pos.z = building.pos.z
        self.pos2.x = building.pos2.x
        self.pos2.y = building.pos2.y
        self.pos2.z = building.pos2.z
        self.yaw = building.yaw
        self.yaw2 = building.yaw2
        self.item_id = building.item_id
        self.model_index = building.model_index
        self.output_object_index = building.output_object_index
        self.input_object_index = building.input_object_index
        self.output_to_slot = building.output_to_slot
        self.input_from_slot = building.input_from_slot
        self.output_from_slot = building.output_from_slot
        self.input_to_slot = building.input_to_slot
        self.output_offset = building.output_offset
        self.input_offset = building.input_offset
        self.recipe_id = building.recipe_id
        self.filter_id = building.filter_id
        self.parameter_count = building.parameter_count
        self.parameters = building.parameters
        
    
    def get_size(self):
        return 61 + 4 * self.parameter_count
    
    def parse(self, packet: Packet):
        start_size = len(packet.data)
        self.index = packet.parse_int32()
        self.area_index = packet.parse_int8()
        self.pos.x = packet.parse_float32()
        self.pos.y = packet.parse_float32()
        self.pos.z = packet.parse_float32()
        self.pos2.x = packet.parse_float32()
        self.pos2.y = packet.parse_float32()
        self.pos2.z = packet.parse_float32()
        self.yaw = packet.parse_float32()
        self.yaw2 = packet.parse_float32()
        self.item_id = packet.parse_int16()
        self.model_index = packet.parse_int16()
        self.output_object_index = packet.parse_int32()
        self.input_object_index = packet.parse_int32()
        self.output_to_slot = packet.parse_int8()
        self.input_from_slot = packet.parse_int8()
        self.output_from_slot = packet.parse_int8()
        self.input_to_slot = packet.parse_int8()
        self.output_offset = packet.parse_int8()
        self.input_offset = packet.parse_int8()
        self.recipe_id = packet.parse_int16()
        self.filter_id = packet.parse_int16()
        self.parameter_count = packet.parse_int16()
        self.parameters = []
        for i in range(self.parameter_count):
            self.parameters.append(packet.parse_int32())
        end_size = len(packet.data)
        assert start_size - end_size == self.get_size(), "Wrong amount of data parsed!"
        
    def serialize(self):
        packet = Packet()
        packet.serialize_int32(self.index)
        packet.serialize_int8(self.area_index)
        packet.serialize_float32(self.pos.x)
        packet.serialize_float32(self.pos.y)
        packet.serialize_float32(self.pos.z)
        packet.serialize_float32(self.pos2.x)
        packet.serialize_float32(self.pos2.y)
        packet.serialize_float32(self.pos2.z)
        packet.serialize_float32(self.yaw)
        packet.serialize_float32(self.yaw2)
        packet.serialize_int16(self.item_id)
        packet.serialize_int16(self.model_index)
        packet.serialize_int32(self.output_object_index)
        packet.serialize_int32(self.input_object_index)
        packet.serialize_int8(self.output_to_slot)
        packet.serialize_int8(self.input_from_slot)
        packet.serialize_int8(self.output_from_slot)
        packet.serialize_int8(self.input_to_slot)
        packet.serialize_int8(self.output_offset)
        packet.serialize_int8(self.input_offset)
        packet.serialize_int16(self.recipe_id)
        packet.serialize_int16(self.filter_id)
        packet.serialize_int16(self.parameter_count)
        for param in self.parameters:
            packet.serialize_int32(param)
        assert len(packet.data) == self.get_size(), "Wrong amount of data serialized!"
        return packet

    def __str__(self):
        packet = self.serialize()
        string = f"""
Blue Print Building:
====================
Version: 1
====================
Length: {len(packet.data)}
Binary data: {packet.data}
====================
Index: {self.index}
Area index: {self.area_index}
position 1 x: {self.pos.x}
position 1 y: {self.pos.y}
position 1 z: {self.pos.z}
position 2 x: {self.pos2.x}
position 2 y: {self.pos2.y}
position 2 z: {self.pos2.z}
Yaw: {self.yaw}
Yaw2: {self.yaw2}
Item ID: {self.item_id} ({BuildingItem(self.item_id).name})
Model index: {self.model_index} ({BuildingModel(self.model_index).name})
Output object index: {self.output_object_index}
Input object index: {self.input_object_index}
Output to slot: {self.output_to_slot}
Input from slot: {self.input_from_slot}
Output from slot: {self.output_from_slot}
Input to slot: {self.input_to_slot}
Output offset: {self.output_offset}
Input offset: {self.input_offset}
Recipe id: {self.recipe_id}
Filter id: {self.filter_id}
Parameter count: {self.parameter_count}
"""
        for i in range(self.parameter_count):
            string += f"\tParam_{i}: {self.parameters[i]}\n"
        return string
    
class BlueprintBuildingV2(BlueprintBuilding):

    def __init__(self, building):
        super().__init__()
        self.dummy1 = -100
        self.dummy2 = 0
        self.index = building.index
        self.area_index = building.area_index
        self.pos.x = building.pos.x
        self.pos.y = building.pos.y
        self.pos.z = building.pos.z
        self.pos2.x = building.pos2.x
        self.pos2.y = building.pos2.y
        self.pos2.z = building.pos2.z
        self.yaw = building.yaw
        self.yaw2 = building.yaw2
        self.item_id = building.item_id
        self.model_index = building.model_index
        self.output_object_index = building.output_object_index
        self.input_object_index = building.input_object_index
        self.output_to_slot = building.output_to_slot
        self.input_from_slot = building.input_from_slot
        self.output_from_slot = building.output_from_slot
        self.input_to_slot = building.input_to_slot
        self.output_offset = building.output_offset
        self.input_offset = building.input_offset
        self.recipe_id = building.recipe_id
        self.filter_id = building.filter_id
        self.parameter_count = building.parameter_count
        self.parameters = building.parameters
    
    def get_size(self):
        return 69 + 4 * self.parameter_count
    
    def parse(self, packet: Packet):
        start_size = len(packet.data)
        self.dummy1 = packet.parse_int32()
        self.index = packet.parse_int32()
        self.area_index = packet.parse_int8()
        self.pos.x = packet.parse_float32()
        self.pos.y = packet.parse_float32()
        self.pos.z = packet.parse_float32()
        self.pos2.x = packet.parse_float32()
        self.pos2.y = packet.parse_float32()
        self.pos2.z = packet.parse_float32()
        self.yaw = packet.parse_float32()
        self.yaw2 = packet.parse_float32()
        self.output_object_index = packet.parse_int32()
        self.item_id = packet.parse_int16()
        self.model_index = packet.parse_int16()
        self.input_object_index = packet.parse_int32()
        self.output_to_slot = packet.parse_int8()
        self.input_from_slot = packet.parse_int8()
        self.output_from_slot = packet.parse_int8()
        self.input_to_slot = packet.parse_int8()
        self.output_offset = packet.parse_int8()
        self.input_offset = packet.parse_int8()
        self.dummy2 = packet.parse_int32()
        self.recipe_id = packet.parse_int16()
        self.filter_id = packet.parse_int16()
        self.parameter_count = packet.parse_int16()
        self.parameters = []
        for i in range(self.parameter_count):
            self.parameters.append(packet.parse_int32())
        end_size = len(packet.data)
        assert start_size - end_size == self.get_size(), "Wrong amount of data parsed!"
        
    def serialize(self):
        packet = Packet()
        packet.serialize_int32(self.dummy1)
        packet.serialize_int32(self.index)
        packet.serialize_int8(self.area_index)
        packet.serialize_float32(self.pos.x)
        packet.serialize_float32(self.pos.y)
        packet.serialize_float32(self.pos.z)
        packet.serialize_float32(self.pos2.x)
        packet.serialize_float32(self.pos2.y)
        packet.serialize_float32(self.pos2.z)
        packet.serialize_float32(self.yaw)
        packet.serialize_float32(self.yaw2)
        packet.serialize_int32(self.output_object_index)
        packet.serialize_int16(self.item_id)
        packet.serialize_int16(self.model_index)
        packet.serialize_int32(self.input_object_index)
        packet.serialize_int8(self.output_to_slot)
        packet.serialize_int8(self.input_from_slot)
        packet.serialize_int8(self.output_from_slot)
        packet.serialize_int8(self.input_to_slot)
        packet.serialize_int8(self.output_offset)
        packet.serialize_int8(self.input_offset)
        packet.serialize_int32(self.dummy2)
        packet.serialize_int16(self.recipe_id)
        packet.serialize_int16(self.filter_id)
        packet.serialize_int16(self.parameter_count)
        for param in self.parameters:
            packet.serialize_int32(param)
        assert len(packet.data) == self.get_size(), "Wrong amount of data serialized!"
        return packet

    def __str__(self):
        packet = self.serialize()
        string = f"""
Blue Print Building:
====================
Version: 2
====================
Length: {len(packet.data)}
Binary data: {packet.data}
====================
Dummy 1: {self.dummy1}
Index: {self.index}
Area index: {self.area_index}
position 1 x: {self.pos.x}
position 1 y: {self.pos.y}
position 1 z: {self.pos.z}
position 2 x: {self.pos2.x}
position 2 y: {self.pos2.y}
position 2 z: {self.pos2.z}
Yaw: {self.yaw}
Yaw2: {self.yaw2}
Output object index: {self.output_object_index}
Item ID: {self.item_id} ({BuildingItem(self.item_id).name})
Model index: {self.model_index} ({BuildingModel(self.model_index).name})
Input object index: {self.input_object_index}
Output to slot: {self.output_to_slot}
Input from slot: {self.input_from_slot}
Output from slot: {self.output_from_slot}
Input to slot: {self.input_to_slot}
Output offset: {self.output_offset}
Input offset: {self.input_offset}
Dummy 2: {self.dummy2}
Recipe id: {self.recipe_id}
Filter id: {self.filter_id}
Parameter count: {self.parameter_count}
"""
        for i in range(self.parameter_count):
            string += f"\tParam_{i}: {self.parameters[i]}\n"
        return string


from .md5 import MD5
from .blueprint_area import BlueprintArea
from .blueprint_building import BlueprintBuilding, BlueprintBuildingV1, BlueprintBuildingV2
from .blueprint_building_header import BlueprintBuildingHeader
from .blueprint_header import BlueprintHeader
from .blueprint_string_header import BlueprintStringHeader
from .packet import Packet
import base64
import gzip

class Blueprint:
    
    def __init__(self):
        pass

    def validate_hash(self):
        calculated_input_hash_value = MD5(MD5.Variant.MD5F).update(self.hashed_string.encode("utf-8")).hexdigest().upper()
        return calculated_input_hash_value == self.hash_str

    def parse(self, blueprint_string, debug = False, debug_raw_data = False):

        HASH_SIZE = 32
        BLUEPRINT_STRING_HEADER_SIZE = 52
        BLUEPRINT_HEADER_SIZE = 29
        BLUEPRINT_AREA_SIZE = 14
        BLUEPRINT_BUILDING_HEADER_SIZE = 4
        BLUEPRINT_BUILDING_SIZE = 61
        
        # Parse and validate MD5 hash
        if len(blueprint_string) < HASH_SIZE:
            return None
        self.hash_str = blueprint_string[-HASH_SIZE:].upper()
        if debug_raw_data:
            print("Hash str:")
            print(self.hash_str)
            print()
        self.data_str = blueprint_string[:-HASH_SIZE]
        self.data_str = self.data_str[:-1] # Remove the extra double quote that is placed to seperate hash and data
        self.hashed_string = self.data_str
        if debug_raw_data:
            print("Data str:")
            print(self.data_str)
            print()
        
        # Parse the blue print string header
        self.data_str = self.data_str.split(",\"")
        if len(self.data_str) != 2:
            return None
        header_str = self.data_str[0]
        if debug_raw_data:
            print("Header str:")
            print(header_str)
            print()
        self.data_str = self.data_str[1]
        if debug_raw_data:
            print("Base64 data str:")
            print(self.data_str)
            print()
        if len(header_str) < BLUEPRINT_STRING_HEADER_SIZE:
            return None
        self.blueprint_string_header = BlueprintStringHeader()
        self.blueprint_string_header.parse(header_str)
        if debug:
            print(self.blueprint_string_header)

        # Base64 decode and gzip decompress the data
        compressed_data = base64.b64decode(self.data_str)
        if debug_raw_data:
            print("Compressed data:")
            print(compressed_data)
            print()
        data = gzip.decompress(compressed_data)
        if debug_raw_data:
            print("Uncompressed data:")
            print(data)
            print()
        packet = Packet(data)

        # Parse the blue print header
        if len(packet.data) < BLUEPRINT_HEADER_SIZE:
            return None
        self.blueprint_header = BlueprintHeader()
        self.blueprint_header.parse(packet)
        if debug:
            print(self.blueprint_header)

        # Parse all areas
        if len(packet.data) < BLUEPRINT_AREA_SIZE * self.blueprint_header.area_count:
            return None
        self.blueprint_areas = []
        for i in range(self.blueprint_header.area_count):
            area = BlueprintArea()
            area.parse(packet)
            self.blueprint_areas.append(area)
            if debug:
                print(area)

        # Parse building header
        if len(packet.data) < BLUEPRINT_BUILDING_HEADER_SIZE:
            return None
        self.blueprint_building_header = BlueprintBuildingHeader()
        self.blueprint_building_header.parse(packet)
        if debug:
            print(self.blueprint_building_header)

        # Parse all buildings
        if len(packet.data) < BLUEPRINT_BUILDING_SIZE * self.blueprint_building_header.building_count:
            return None
        self.blueprint_buildings = []
        for i in range(self.blueprint_building_header.building_count):
            blueprint_building = BlueprintBuilding()
            blueprint_building = BlueprintBuilding.get_version(packet)(blueprint_building)
            blueprint_building.parse(packet)
            self.blueprint_buildings.append(blueprint_building)
            if debug:
                print(blueprint_building)

        return self.blueprint_buildings

    def serialize(self, blueprint_buildings, debug = False, debug_raw_data = False, blueprint_building_version = BlueprintBuildingV1):

        data = bytes()

        # Serialize buildings
        blueprint_buildings_data = bytes()
        for blueprint_building in blueprint_buildings:
            blueprint_building = blueprint_building_version(blueprint_building)
            building_data = blueprint_building.serialize().data
            blueprint_buildings_data += building_data
            if debug:
                print("Blueprint building:")
                print(blueprint_building)
                print()
        data = blueprint_buildings_data + data

        # Serialize building header
        self.blueprint_building_header = BlueprintBuildingHeader(len(blueprint_buildings))
        blueprint_building_header_data = self.blueprint_building_header.serialize().data
        data = blueprint_building_header_data + data
        if debug:
            print("Blueprint data:", blueprint_building_header_data)

        # Serialize areas
        size, offset = BlueprintArea.get_area_from_building_list(blueprint_buildings)
        self.blueprint_areas = [BlueprintArea(size = size, offset = offset)]
        blueprint_areas_data = bytes()
        for blueprint_area in self.blueprint_areas:
            blueprint_areas_data += blueprint_area.serialize().data
            if debug:
                print("Blueprint areas data:", blueprint_areas_data)
        data = blueprint_areas_data + data

        # Serialize blueprint header
        self.blueprint_header = BlueprintHeader(size)
        blueprint_header_data = self.blueprint_header.serialize().data
        data = blueprint_header_data + data
        if debug:
            print("Blueprint header data:", blueprint_header_data)
        if debug_raw_data:
            print("Uncompressed data:")
            print(data)
            print()

        # Zip compress and base64 encode the data
        compressed_data = gzip.compress(data, compresslevel = 8, mtime = 0)
        if debug_raw_data:
            print("Compressed data:")
            print(compressed_data)
            print()
        temp = bytearray(compressed_data)
        temp.append(0x00)
        temp[9] = 0x0b
        compressed_data = bytes(temp)

        data_str = base64.b64encode(compressed_data).decode('utf-8')
        if debug_raw_data:
            print("Base64 data str:")
            print(data_str)
            print()

        # Append blueprint string header
        self.blueprint_string_header = BlueprintStringHeader()
        blueprint_string_header_str = self.blueprint_string_header.serialize()
        data_str = blueprint_string_header_str + "\"" + data_str
        if debug_raw_data:
            print("Header str:")
            print(blueprint_string_header_str)
            print()
        
        # Calculate and append md5 sum
        calculated_input_hash_value = MD5(MD5.Variant.MD5F).update(data_str.encode("utf-8")).hexdigest().upper()
        if debug_raw_data:
            print("Hash str:")
            print(calculated_input_hash_value)
            print()
        data_str += "\"" + calculated_input_hash_value

        # Return blueprint string
        return data_str


from .building import Factory3x3
from ..utils import Vector, Yaw
from ..enums import BuildingItem, BuildingModel

class Smelter(Factory3x3):
    def __init__(self, name: str, pos: Vector, recipe_id: int = 0):
        super().__init__(name)
        self.pos = pos
        self.pos2 = pos
        self.yaw = Yaw.North
        self.yaw2 = Yaw.North
        self.output_object_index = -1
        self.input_object_index = -1
        self.output_to_slot = 0
        self.input_from_slot = 0
        self.output_from_slot = 0
        self.input_to_slot = 0
        self.recipe_id = recipe_id
        self.parameter_count = 1
        self.parameters = [0]
        
    def get_size():
        return Vector(3.0, 3.0)
    
    def get_offset():
        return Vector(1.5, 1.5)
        
class ArcSmelter(Smelter):
    def __init__(self, name: str, pos: Vector, recipe_id: int = 0):
        super().__init__(name, pos, recipe_id)
        self.item_id = BuildingItem.ArcSmelter
        self.model_index = BuildingModel.ArcSmelter
        
class PlaneSmelter(Smelter):
    def __init__(self, name: str, pos: Vector, recipe_id: int = 0):
        super().__init__(name, pos, recipe_id)
        self.item_id = BuildingItem.PlaneSmelter
        self.model_index = BuildingModel.PlaneSmelter
        
class NegentrophySmelter(Smelter):
    def __init__(self, name: str, pos: Vector, recipe_id: int = 0):
        super().__init__(name, pos, recipe_id)
        self.item_id = BuildingItem.NegentrophySmelter
        self.model_index = BuildingModel.NegentrophySmelter



from ..blueprint import BlueprintBuilding
from ..utils import Vector, Yaw
from ..enums import BuildingItem, BuildingModel

class Building(BlueprintBuilding):
    
    buildings = []
    
    def __init__(self, name = "Unknown", **kwargs):
        super().__init__(**kwargs)
        self.__class__.buildings.append(self)
        self.index = len(Building.buildings) - 1
        self.name = name
    
    def count():
        return len(Building.buildings)
    
    def get_last_building():
        return Building.buildings[-1]
    
    def get_building(index):
        return Building.buildings[index]
    
    def get_nearest_slot_from_position(self, pos):
        distances = []
        for i in range(self.number_of_slots()):
            distance = pos.get_distance(self.get_position_of_slot(i))
            distances.append(distance)
        slot = distances.index(min(distances))
        return slot
    
    def __str__(self):
        string = f"""
Blue Print Building:
====================
Name: {self.name}
====================
Index: {self.index}
Area index: {self.area_index}
position 1 x: {self.pos.x}
position 1 y: {self.pos.y}
position 1 z: {self.pos.z}
position 2 x: {self.pos2.x}
position 2 y: {self.pos2.y}
position 2 z: {self.pos2.z}
Yaw: {self.yaw}
Yaw2: {self.yaw2}
Item ID: {str(BuildingItem(self.item_id))} ({self.item_id})
Model index: {str(BuildingModel(self.model_index))} ({self.model_index})
Output object index: {self.output_object_index}
Input object index: {self.input_object_index}
Output to slot: {self.output_to_slot}
Input from slot: {self.input_from_slot}
Output from slot: {self.output_from_slot}
Input to slot: {self.input_to_slot}
Output offset: {self.output_offset}
Input offset: {self.input_offset}
Recipe id: {self.recipe_id}
Filter id: {self.filter_id}
Parameter count: {self.parameter_count}
"""
        for i in range(self.parameter_count):
            string += f"\tParam_{i}: {self.parameters[i]}\n"
        return string

class Factory(Building):
    
    def __init__(self, name = "Unknown", **kwargs):
        super().__init__(name, **kwargs)

    def set_recipe(self, recipe_id):
        self.recipe_id = recipe_id

class Factory3x3(Factory):

#          slot 0  slot 1  slot 2
#         ┌───────────────────────┐
#         │                       │
# slot 11 │                       │ slot 3
#         │                       │
#         │                       │
# slot 10 │           X           │ slot 4
#         │                       │
#         │                       │
# slot 9  │                       │ slot 5
#         │                       │
#         └───────────────────────┘
#           slot 8  slot 7  slot6  
    
    def number_of_slots(self):
        return 12
    
    def __init__(self, name = "Unknown", **kwargs):
        super().__init__(name, **kwargs)
        
    def get_position_of_slot(self, slot):
        assert slot >= 0 and slot <= 11, f"slot index needs to be: slot >= 0 and slot <= 11 (slot was {slot})"
        if slot == 0:
            delta_pos = Vector(x = -0.75, y = 0.8)
        elif slot == 1:
            delta_pos = Vector(x = 0.0, y = 0.8)
        elif slot == 2:
            delta_pos = Vector(x = 0.75, y = 0.8)
        elif slot == 3:
            delta_pos = Vector(x = 0.8, y = 0.75)
        elif slot == 4:
            delta_pos = Vector(x = 0.8, y = 0.0)
        elif slot == 5:
            delta_pos = Vector(x = 0.8, y = -0.75)
        elif slot == 6:
            delta_pos = Vector(x = 0.75, y = -0.8)
        elif slot == 7:
            delta_pos = Vector(x = 0.0, y = -0.8)
        elif slot == 8:
            delta_pos = Vector(x = -0.75, y = -0.8)
        elif slot == 9:
            delta_pos = Vector(x = -0.8, y = -0.75)
        elif slot == 10:
            delta_pos = Vector(x = -0.8, y = 0.0)
        elif slot == 11:
            delta_pos = Vector(x = -0.8, y = 0.75)

        return self.pos + delta_pos


from .building import Building

class MatrixLab(Building):
    pass

    def get_size():
        return Vector(5.0, 5.0)

    def get_offset():
        return Vector(2.5, 2.5)

class SelfEvolutionLab(Building):
    pass


from .packet import Packet
from ..utils import Vector
import sys

class BlueprintArea:

    def __init__(self, size: Vector = Vector(), offset: Vector = Vector()):
        self.index = 0
        self.parent_index = -1
        self.tropic_anchor = 0
        self.area_segments = 200
        self.anchor_local_offset_x = 0
        self.anchor_local_offset_y = 0
        self.width = int(size.x)
        self.height = int(size.y)

    def parse(self, packet):
        self.index = packet.parse_int8()
        self.parent_index = packet.parse_int8()
        self.tropic_anchor = packet.parse_int16()
        self.area_segments = packet.parse_int16()
        self.anchor_local_offset_x = packet.parse_int16()
        self.anchor_local_offset_y = packet.parse_int16()
        self.width = packet.parse_int16()
        self.height = packet.parse_int16()

    def serialize(self):
        packet = Packet()
        packet.serialize_int8(self.index)
        packet.serialize_int8(self.parent_index)
        packet.serialize_int16(self.tropic_anchor)
        packet.serialize_int16(self.area_segments)
        packet.serialize_int16(self.anchor_local_offset_x)
        packet.serialize_int16(self.anchor_local_offset_y)
        packet.serialize_int16(self.width)
        packet.serialize_int16(self.height)
        return packet
    
    def get_area_from_building_list(buildings):
        min_x = 0x7fff
        max_x = -0x7fff
        min_y = 0x7fff
        max_y = -0x7fff
        for building in buildings:
            min_x = min(min_x, min(building.pos.x, building.pos2.x))
            max_x = max(max_x, max(building.pos.x, building.pos2.x))
            min_y = min(min_y, min(building.pos.y, building.pos2.y))
            max_y = max(max_y, max(building.pos.y, building.pos2.y))
        size = Vector(max_x - min_x + 1, max_y - min_y + 1)
        offset = Vector((max_x - min_x) // 2, (max_y - min_y) // 2)
        return size, offset

    def __str__(self):
        return f"""
Blueprint Area:
===============
Binary data: {self.serialize().data}
===============
Index: {self.index}
Parent index: {self.parent_index}
Tropic anchor: {self.tropic_anchor}
Area segments: {self.area_segments}
Anchor local offset x: {self.anchor_local_offset_x}
Anchor local offset Y: {self.anchor_local_offset_y}
Width: {self.width}
Height: {self.height}
"""


from .recipes import Recipe
from ..enums import Item
from .factory_section import FactorySection
from ..buildings import ArcSmelter, AssemblingMachineMKI, OilRefinary, ChemicalPlant, MatrixLab
from ..utils import Vector
from math import ceil

class Factory:

    def __init__(self):
        self.prolifirator = None # Mk.I, Mk.II, Mk.III
        self.input_flow = []

    def set_tartget_output_flow(self, products, debug = False):

        requirement_stack = products
        production_stack = []

        while len(requirement_stack) > 0:
            
            item = requirement_stack.pop()
            
            # Add ingredients to stack
            for ingredient in item.get_needed_ingredients(self.prolifirator):
                requirement_stack.append(ingredient)

            # Add item to production stack
            index = -1
            for i in range(len(production_stack)):
                if production_stack[i].name == item.name:
                    index = i
                    break
            if index != -1:
                ingredient_to_bump = production_stack.pop(index)
                ingredient_to_bump.count_pr_sec += item.count_pr_sec
                production_stack.append(ingredient_to_bump)
            else:
                production_stack.append(item)

        self.target_output_flow = production_stack[::-1] # Reverse list
        self.input_flow = []
        for flow in self.target_output_flow:
            if Recipe.recipes[flow.name] == None:
                self.input_flow.append(flow)
                self.target_output_flow.remove(flow)

        if debug:
            print("Inputs stack")
            for item in self.input_flow:
                print(f"\t{item.name}: {float(item.count_pr_sec)}/s")
            print("Outputs stack")
            for item in self.target_output_flow:
                print(f"\t{item.name}: {float(item.count_pr_sec)}/s")

    def generate_factories(self, debug = False):
        
        # Define main belt
        self.main_belts = []
        for item in self.input_flow:
            self.main_belts.append(item)
        for item in self.target_output_flow:
            self.main_belts.append(item)
        
        # Print debug for main belt
        if debug:
            print("Main belts:")
            for belt in self.main_belts:
                print(f"\t{belt.name}, {belt.count_pr_sec}/s")


        self.factories = []
        input_count = len(self.input_flow)
        output_count = 0
        y = 0
        
        for product in self.target_output_flow:
            
            recipe = Recipe.select(product.name)

            # If the product is a raw material, skip it
            if Recipe.recipes[product.name] == None:
                continue
            self.factories.append(
                FactorySection(
                    pos = Vector(0, y),
                    input_count = input_count,
                    output_count = output_count,
                    main_belts = self.main_belts,
                    product = product,
                    recipe = recipe
                )
            )
            
            if len(self.factories) > 1:
                self.factories[-2].connect_to_section(self.factories[-1])
            
            output_count += self.factories[-1].product_count
            y += self.factories[-1].height

    def get_factory(self, factory_type):
        factories = {
            "Smelting Facility": ArcSmelter,
            "Assembling machine": AssemblingMachineMKI,
            "Refining Facility": OilRefinary,
            "Chemical Facility": ChemicalPlant,
            "Matrix Lab": MatrixLab
        }
        return factories[factory_type]

    def generate_blueprint(self):
        pass



import re
import math
import collections
import enum

class MD5():
	class Variant(enum.IntEnum):
		Original = 0
		MD5F = 1
		MD5FC = 2

	def _f(x, y, z):
		return (x & y) | (~x & z)

	def _g(x, y, z):
		return (x & z) | (y & ~z)

	def _h(x, y, z):
		return x ^ y ^ z

	def _i(x, y, z):
		return y ^ (x | ~z)

	@staticmethod
	def _rol(x, s):
		return ((x << s) | (x >> (32 - s))) & 0xffffffff

	_RoundOp = collections.namedtuple("RoundOp", [ "a", "b", "c", "d", "k", "s", "i", "T", "op" ])
	_ROUND_OPS = [
		_RoundOp(a = 0, b = 1, c = 2, d = 3, k =  0, s =  7, i =  1, T = 0xd76aa478, op = _f),
		_RoundOp(a = 3, b = 0, c = 1, d = 2, k =  1, s = 12, i =  2, T = 0xe8c7b756, op = _f),
		_RoundOp(a = 2, b = 3, c = 0, d = 1, k =  2, s = 17, i =  3, T = 0x242070db, op = _f),
		_RoundOp(a = 1, b = 2, c = 3, d = 0, k =  3, s = 22, i =  4, T = 0xc1bdceee, op = _f),
		_RoundOp(a = 0, b = 1, c = 2, d = 3, k =  4, s =  7, i =  5, T = 0xf57c0faf, op = _f),
		_RoundOp(a = 3, b = 0, c = 1, d = 2, k =  5, s = 12, i =  6, T = 0x4787c62a, op = _f),
		_RoundOp(a = 2, b = 3, c = 0, d = 1, k =  6, s = 17, i =  7, T = 0xa8304613, op = _f),
		_RoundOp(a = 1, b = 2, c = 3, d = 0, k =  7, s = 22, i =  8, T = 0xfd469501, op = _f),
		_RoundOp(a = 0, b = 1, c = 2, d = 3, k =  8, s =  7, i =  9, T = 0x698098d8, op = _f),
		_RoundOp(a = 3, b = 0, c = 1, d = 2, k =  9, s = 12, i = 10, T = 0x8b44f7af, op = _f),
		_RoundOp(a = 2, b = 3, c = 0, d = 1, k = 10, s = 17, i = 11, T = 0xffff5bb1, op = _f),
		_RoundOp(a = 1, b = 2, c = 3, d = 0, k = 11, s = 22, i = 12, T = 0x895cd7be, op = _f),
		_RoundOp(a = 0, b = 1, c = 2, d = 3, k = 12, s =  7, i = 13, T = 0x6b901122, op = _f),
		_RoundOp(a = 3, b = 0, c = 1, d = 2, k = 13, s = 12, i = 14, T = 0xfd987193, op = _f),
		_RoundOp(a = 2, b = 3, c = 0, d = 1, k = 14, s = 17, i = 15, T = 0xa679438e, op = _f),
		_RoundOp(a = 1, b = 2, c = 3, d = 0, k = 15, s = 22, i = 16, T = 0x49b40821, op = _f),
		_RoundOp(a = 0, b = 1, c = 2, d = 3, k =  1, s =  5, i = 17, T = 0xf61e2562, op = _g),
		_RoundOp(a = 3, b = 0, c = 1, d = 2, k =  6, s =  9, i = 18, T = 0xc040b340, op = _g),
		_RoundOp(a = 2, b = 3, c = 0, d = 1, k = 11, s = 14, i = 19, T = 0x265e5a51, op = _g),
		_RoundOp(a = 1, b = 2, c = 3, d = 0, k =  0, s = 20, i = 20, T = 0xe9b6c7aa, op = _g),
		_RoundOp(a = 0, b = 1, c = 2, d = 3, k =  5, s =  5, i = 21, T = 0xd62f105d, op = _g),
		_RoundOp(a = 3, b = 0, c = 1, d = 2, k = 10, s =  9, i = 22, T = 0x2441453, op = _g),
		_RoundOp(a = 2, b = 3, c = 0, d = 1, k = 15, s = 14, i = 23, T = 0xd8a1e681, op = _g),
		_RoundOp(a = 1, b = 2, c = 3, d = 0, k =  4, s = 20, i = 24, T = 0xe7d3fbc8, op = _g),
		_RoundOp(a = 0, b = 1, c = 2, d = 3, k =  9, s =  5, i = 25, T = 0x21e1cde6, op = _g),
		_RoundOp(a = 3, b = 0, c = 1, d = 2, k = 14, s =  9, i = 26, T = 0xc33707d6, op = _g),
		_RoundOp(a = 2, b = 3, c = 0, d = 1, k =  3, s = 14, i = 27, T = 0xf4d50d87, op = _g),
		_RoundOp(a = 1, b = 2, c = 3, d = 0, k =  8, s = 20, i = 28, T = 0x455a14ed, op = _g),
		_RoundOp(a = 0, b = 1, c = 2, d = 3, k = 13, s =  5, i = 29, T = 0xa9e3e905, op = _g),
		_RoundOp(a = 3, b = 0, c = 1, d = 2, k =  2, s =  9, i = 30, T = 0xfcefa3f8, op = _g),
		_RoundOp(a = 2, b = 3, c = 0, d = 1, k =  7, s = 14, i = 31, T = 0x676f02d9, op = _g),
		_RoundOp(a = 1, b = 2, c = 3, d = 0, k = 12, s = 20, i = 32, T = 0x8d2a4c8a, op = _g),
		_RoundOp(a = 0, b = 1, c = 2, d = 3, k =  5, s =  4, i = 33, T = 0xfffa3942, op = _h),
		_RoundOp(a = 3, b = 0, c = 1, d = 2, k =  8, s = 11, i = 34, T = 0x8771f681, op = _h),
		_RoundOp(a = 2, b = 3, c = 0, d = 1, k = 11, s = 16, i = 35, T = 0x6d9d6122, op = _h),
		_RoundOp(a = 1, b = 2, c = 3, d = 0, k = 14, s = 23, i = 36, T = 0xfde5380c, op = _h),
		_RoundOp(a = 0, b = 1, c = 2, d = 3, k =  1, s =  4, i = 37, T = 0xa4beea44, op = _h),
		_RoundOp(a = 3, b = 0, c = 1, d = 2, k =  4, s = 11, i = 38, T = 0x4bdecfa9, op = _h),
		_RoundOp(a = 2, b = 3, c = 0, d = 1, k =  7, s = 16, i = 39, T = 0xf6bb4b60, op = _h),
		_RoundOp(a = 1, b = 2, c = 3, d = 0, k = 10, s = 23, i = 40, T = 0xbebfbc70, op = _h),
		_RoundOp(a = 0, b = 1, c = 2, d = 3, k = 13, s =  4, i = 41, T = 0x289b7ec6, op = _h),
		_RoundOp(a = 3, b = 0, c = 1, d = 2, k =  0, s = 11, i = 42, T = 0xeaa127fa, op = _h),
		_RoundOp(a = 2, b = 3, c = 0, d = 1, k =  3, s = 16, i = 43, T = 0xd4ef3085, op = _h),
		_RoundOp(a = 1, b = 2, c = 3, d = 0, k =  6, s = 23, i = 44, T = 0x4881d05, op = _h),
		_RoundOp(a = 0, b = 1, c = 2, d = 3, k =  9, s =  4, i = 45, T = 0xd9d4d039, op = _h),
		_RoundOp(a = 3, b = 0, c = 1, d = 2, k = 12, s = 11, i = 46, T = 0xe6db99e5, op = _h),
		_RoundOp(a = 2, b = 3, c = 0, d = 1, k = 15, s = 16, i = 47, T = 0x1fa27cf8, op = _h),
		_RoundOp(a = 1, b = 2, c = 3, d = 0, k =  2, s = 23, i = 48, T = 0xc4ac5665, op = _h),
		_RoundOp(a = 0, b = 1, c = 2, d = 3, k =  0, s =  6, i = 49, T = 0xf4292244, op = _i),
		_RoundOp(a = 3, b = 0, c = 1, d = 2, k =  7, s = 10, i = 50, T = 0x432aff97, op = _i),
		_RoundOp(a = 2, b = 3, c = 0, d = 1, k = 14, s = 15, i = 51, T = 0xab9423a7, op = _i),
		_RoundOp(a = 1, b = 2, c = 3, d = 0, k =  5, s = 21, i = 52, T = 0xfc93a039, op = _i),
		_RoundOp(a = 0, b = 1, c = 2, d = 3, k = 12, s =  6, i = 53, T = 0x655b59c3, op = _i),
		_RoundOp(a = 3, b = 0, c = 1, d = 2, k =  3, s = 10, i = 54, T = 0x8f0ccc92, op = _i),
		_RoundOp(a = 2, b = 3, c = 0, d = 1, k = 10, s = 15, i = 55, T = 0xffeff47d, op = _i),
		_RoundOp(a = 1, b = 2, c = 3, d = 0, k =  1, s = 21, i = 56, T = 0x85845dd1, op = _i),
		_RoundOp(a = 0, b = 1, c = 2, d = 3, k =  8, s =  6, i = 57, T = 0x6fa87e4f, op = _i),
		_RoundOp(a = 3, b = 0, c = 1, d = 2, k = 15, s = 10, i = 58, T = 0xfe2ce6e0, op = _i),
		_RoundOp(a = 2, b = 3, c = 0, d = 1, k =  6, s = 15, i = 59, T = 0xa3014314, op = _i),
		_RoundOp(a = 1, b = 2, c = 3, d = 0, k = 13, s = 21, i = 60, T = 0x4e0811a1, op = _i),
		_RoundOp(a = 0, b = 1, c = 2, d = 3, k =  4, s =  6, i = 61, T = 0xf7537e82, op = _i),
		_RoundOp(a = 3, b = 0, c = 1, d = 2, k = 11, s = 10, i = 62, T = 0xbd3af235, op = _i),
		_RoundOp(a = 2, b = 3, c = 0, d = 1, k =  2, s = 15, i = 63, T = 0x2ad7d2bb, op = _i),
		_RoundOp(a = 1, b = 2, c = 3, d = 0, k =  9, s = 21, i = 64, T = 0xeb86d391, op = _i),
	]
	_ROUND_OP_PATCHES = {
		Variant.MD5F: {
			1: _RoundOp(a = 3, b = 0, c = 1, d = 2, k =  1, s = 12, i =  2, T = 0xe8d7b756, op = _f),
			6: _RoundOp(a = 2, b = 3, c = 0, d = 1, k =  6, s = 17, i =  7, T = 0xa8304623, op = _f),
			12: _RoundOp(a = 0, b = 1, c = 2, d = 3, k = 12, s =  7, i = 13, T = 0x6b9f1122, op = _f),
			15: _RoundOp(a = 1, b = 2, c = 3, d = 0, k = 15, s = 22, i = 16, T = 0x39b40821, op = _f),
			19: _RoundOp(a = 1, b = 2, c = 3, d = 0, k =  0, s = 20, i = 20, T = 0xc9b6c7aa, op = _g),
			21: _RoundOp(a = 3, b = 0, c = 1, d = 2, k = 10, s =  9, i = 22, T = 0x2443453, op = _g),
			24: _RoundOp(a = 0, b = 1, c = 2, d = 3, k =  9, s =  5, i = 25, T = 0x21f1cde6, op = _g),
			27: _RoundOp(a = 1, b = 2, c = 3, d = 0, k =  8, s = 20, i = 28, T = 0x475a14ed, op = _g),
		},
		Variant.MD5FC: {
			1: _RoundOp(a = 3, b = 0, c = 1, d = 2, k =  1, s = 12, i =  2, T = 0xe8d7b756, op = _f),
			3: _RoundOp(a = 1, b = 2, c = 3, d = 0, k =  3, s = 22, i =  4, T = 0xc1bdceef, op = _f),
			6: _RoundOp(a = 2, b = 3, c = 0, d = 1, k =  6, s = 17, i =  7, T = 0xa8304623, op = _f),
			12: _RoundOp(a = 0, b = 1, c = 2, d = 3, k = 12, s =  7, i = 13, T = 0x6b9f1122, op = _f),
			15: _RoundOp(a = 1, b = 2, c = 3, d = 0, k = 15, s = 22, i = 16, T = 0x39b40821, op = _f),
			19: _RoundOp(a = 1, b = 2, c = 3, d = 0, k =  0, s = 20, i = 20, T = 0xc9b6c7aa, op = _g),
			21: _RoundOp(a = 3, b = 0, c = 1, d = 2, k = 10, s =  9, i = 22, T = 0x2443453, op = _g),
			24: _RoundOp(a = 0, b = 1, c = 2, d = 3, k =  9, s =  5, i = 25, T = 0x23f1cde6, op = _g),
			27: _RoundOp(a = 1, b = 2, c = 3, d = 0, k =  8, s = 20, i = 28, T = 0x475a14ed, op = _g),
			34: _RoundOp(a = 2, b = 3, c = 0, d = 1, k = 11, s = 16, i = 35, T = 0x6d9d6121, op = _h),
		},
	}
	_INIT_VALUES = {
		Variant.Original: (
			int.from_bytes(bytes.fromhex("01 23 45 67"), byteorder = "little"),
			int.from_bytes(bytes.fromhex("89 ab cd ef"), byteorder = "little"),
			int.from_bytes(bytes.fromhex("fe dc ba 98"), byteorder = "little"),
			int.from_bytes(bytes.fromhex("76 54 32 10"), byteorder = "little"),
		),
		Variant.MD5F: (
			int.from_bytes(bytes.fromhex("01 23 45 67"), byteorder = "little"),
			int.from_bytes(bytes.fromhex("89 ab dc ef"), byteorder = "little"),
			int.from_bytes(bytes.fromhex("fe dc ba 98"), byteorder = "little"),
			int.from_bytes(bytes.fromhex("46 57 32 10"), byteorder = "little"),
		),
		Variant.MD5FC: (
			int.from_bytes(bytes.fromhex("01 23 45 67"), byteorder = "little"),
			int.from_bytes(bytes.fromhex("89 ab dc ef"), byteorder = "little"),
			int.from_bytes(bytes.fromhex("fe dc ba 98"), byteorder = "little"),
			int.from_bytes(bytes.fromhex("46 57 32 10"), byteorder = "little"),
		),
	}

	def __init__(self, variant = Variant.Original):
		(self._a, self._b, self._c, self._d) = self._INIT_VALUES[variant]
		self._buffer = bytearray()
		self._digest = None
		self._length = 0
		self._patches = self._ROUND_OP_PATCHES.get(variant, { })

	def _update_block(self, block):
		assert(len(block) == 64)
		state = [ self._a, self._b, self._c, self._d ]
		x = [ int.from_bytes(block[4 * i : 4 * (i + 1)], byteorder = "little") for i in range(16) ]

		for (i, round_op) in enumerate(self._ROUND_OPS):
			if i in self._patches:
				round_op = self._patches[i]
			a = state[round_op.a]
			b = state[round_op.b]
			c = state[round_op.c]
			d = state[round_op.d]
			state[round_op.a] = b + self._rol((a + round_op.op(b, c, d) + x[round_op.k] + round_op.T) & 0xffffffff, round_op.s)
		self._a = (self._a + state[0]) & 0xffffffff
		self._b = (self._b + state[1]) & 0xffffffff
		self._c = (self._c + state[2]) & 0xffffffff
		self._d = (self._d + state[3]) & 0xffffffff

	def _update(self, data, count_length = True):
		assert(self._digest is None)
		if count_length:
			self._length += len(data)
		self._buffer += data
		while len(self._buffer) >= 64:
			block = self._buffer[:64]
			self._buffer = self._buffer[64:]
			self._update_block(block)
		return self

	def update(self, data):
		return self._update(data)

	def _finalize(self):
		if self._digest is not None:
			return

		# Add padding
		padding_len = (64 - (self._length % 64) - 8) % 64
		if padding_len == 0:
			padding_len = 64
		self._update(bytes([ 0x80 ]), count_length = False)
		self._update(bytes([ 0 ] * (padding_len - 1)), count_length = False)

		# Add length
		length_bits = int.to_bytes(self._length * 8, length = 8, byteorder = "little")
		self._update(length_bits, count_length = False)

		# Concatenate digest
		self._digest = b"".join(int.to_bytes(x, length = 4, byteorder = "little") for x in [ self._a, self._b, self._c, self._d ])

	def digest(self):
		self._finalize()
		return self._digest

	def hexdigest(self):
		return self.digest().hex()

	@classmethod
	def _generate_block(cls, op, round_text):
		regex = re.compile(r"\[(?P<order>[A-D]{4})\s+(?P<k>\d+)\s+(?P<s>\d+)\s+(?P<i>\d+)\]")
		for match in regex.finditer(round_text):
			match = match.groupdict()
			a = "ABCD".index(match["order"][0])
			b = "ABCD".index(match["order"][1])
			c = "ABCD".index(match["order"][2])
			d = "ABCD".index(match["order"][3])
			k = int(match["k"])
			s = int(match["s"])
			i = int(match["i"])
			T = int(abs(math.sin(i)) * 0x100000000) & 0xffffffff
			print("\t\t_RoundOp(a = %d, b = %d, c = %d, d = %d, k = %2d, s = %2d, i = %2d, T = 0x%x, op = _%s)," % (a, b, c, d, k, s, i, T, op))

	@classmethod
	def generate(cls):
		cls._generate_block(op = "f", round_text = """
			[ABCD  0  7  1]  [DABC  1 12  2]  [CDAB  2 17  3]  [BCDA  3 22  4]
			[ABCD  4  7  5]  [DABC  5 12  6]  [CDAB  6 17  7]  [BCDA  7 22  8]
			[ABCD  8  7  9]  [DABC  9 12 10]  [CDAB 10 17 11]  [BCDA 11 22 12]
			[ABCD 12  7 13]  [DABC 13 12 14]  [CDAB 14 17 15]  [BCDA 15 22 16]
		""")
		cls._generate_block(op = "g", round_text = """
			[ABCD  1  5 17]  [DABC  6  9 18]  [CDAB 11 14 19]  [BCDA  0 20 20]
			[ABCD  5  5 21]  [DABC 10  9 22]  [CDAB 15 14 23]  [BCDA  4 20 24]
			[ABCD  9  5 25]  [DABC 14  9 26]  [CDAB  3 14 27]  [BCDA  8 20 28]
			[ABCD 13  5 29]  [DABC  2  9 30]  [CDAB  7 14 31]  [BCDA 12 20 32]
		""")
		cls._generate_block(op = "h", round_text = """
			[ABCD  5  4 33]  [DABC  8 11 34]  [CDAB 11 16 35]  [BCDA 14 23 36]
			[ABCD  1  4 37]  [DABC  4 11 38]  [CDAB  7 16 39]  [BCDA 10 23 40]
			[ABCD 13  4 41]  [DABC  0 11 42]  [CDAB  3 16 43]  [BCDA  6 23 44]
			[ABCD  9  4 45]  [DABC 12 11 46]  [CDAB 15 16 47]  [BCDA  2 23 48]
		""")
		cls._generate_block(op = "i", round_text = """
			[ABCD  0  6 49]  [DABC  7 10 50]  [CDAB 14 15 51]  [BCDA  5 21 52]
			[ABCD 12  6 53]  [DABC  3 10 54]  [CDAB 10 15 55]  [BCDA  1 21 56]
			[ABCD  8  6 57]  [DABC 15 10 58]  [CDAB  6 15 59]  [BCDA 13 21 60]
			[ABCD  4  6 61]  [DABC 11 10 62]  [CDAB  2 15 63]  [BCDA  9 21 64]
		""")

if __name__ == "__main__":
	import os
	import hashlib
	import sys

	if len(sys.argv) == 1:
		for i in range(200):
			data = os.urandom(i)
			right = hashlib.md5(data).hexdigest()
			mine = MD5().update(data).hexdigest()
			if right != mine:
				print(data.hex(), right, mine)
				raise Exception("Failed test cases.")

		assert(MD5(variant = MD5.Variant.MD5F).update(b"").hexdigest() == "84d1ce3bd68f49ab26eb0f96416617cf")
		assert(MD5(variant = MD5.Variant.MD5F).update(b"a").hexdigest() == "f10bddaecb62e5a92433757867ee06db")
		assert(MD5(variant = MD5.Variant.MD5F).update(b"abcd").hexdigest() == "fa27c78b6ec31559f0e760ce3f2b03f6")
		assert(MD5(variant = MD5.Variant.MD5F).update(b"Why are you doing this, Youthcat Studio?").hexdigest() == "13424e12890a3f50a1f8567c464fff8c")
		print("Passed testcases.")
	else:
		md = MD5(variant = MD5.Variant.MD5F)
		with open(sys.argv[1], "rb") as f:
			md.update(f.read())
		print(md.hexdigest())



from .building import Factory
from ..utils import Vector, Yaw

class ChemicalPlant(Factory):
    def __init__(self, name, pos: Vector, yaw: Yaw, recipe_id: int = 0):
        super().__init__(name)
        self.pos = pos
        self.pos2 = pos
        self.yaw = yaw
        self.yaw2 = yaw
        self.item_id = BuildingItem.ChemicalPlant
        self.model_index = BuildingModel.ChemicalPlant
        self.output_object_index = -1
        self.input_object_index = -1
        self.recipe_id = recipe_id
        self.parameter_count = 1
        self.parameters = [0]

    def get_slot_from_pos(self, pos):
        dx = pos.x - self.pos.x
        dy = pos.y - self.pos.y

        if dx < -0.5:
            if dy < -0.5:
                if -dy > -dx:
                    return 8
                else:
                    return 9
            elif dy > 0.5:
                if -dx > dy:
                    return 11
                else:
                    return 0               
            else:
                return 10        
        elif dx > 0.5:
            if dy < -0.5:
                if -dy > dx:
                    return 6
                else:
                    return 5
            elif dy > 0.5:
                if dx > dy:
                    return 3
                else:
                    return 2
            else:
                return 4
        else:
            if dy < -0.5:
                return 7
            elif dy > 0.5:
                return 1
            else:
                assert False, "Can't find slot index when source on top of target"
            
    def get_pos_from_slot(self, slot):
        assert slot >= 0 and slot <= 7, f"slot index needs to be: slot >= 0 and slot <= 7 (slot was {slot})"
        assert False, "Not tested"
        if slot == 0:
            delta_pos = Vector(x = 0.8, y = -0.8)
        elif slot == 1:
            delta_pos = Vector(x = 0.8, y = 0.0)
        elif slot == 2:
            delta_pos = Vector(x = 0.8, y = 0.8)
        elif slot == 3:
            delta_pos = Vector(x = -0.8, y = 0.8)
        elif slot == 4:
            delta_pos = Vector(x = -0.8, y = 0.0)
        elif slot == 5:
            delta_pos = Vector(x = -0.8, y = -0.8)
        elif slot == 6:
            delta_pos = Vector(x = 0.8, y = -0.8)
        elif slot == 7:
            delta_pos = Vector(x = 0.0, y = -0.8)
        elif slot == 8:
            delta_pos = Vector(x = -0.8, y = -0.8)
        elif slot == 9:
            delta_pos = Vector(x = -1.2, y = -0.8)
        elif slot == 10:
            delta_pos = Vector(x = -1.2, y = 0.0)
        elif slot == 11:
            delta_pos = Vector(x = -1.2, y = 0.8)

class QuantumChemicalPlant(ChemicalPlant):
    def __init__(self, name, pos: Vector, yaw: Yaw, recipe_id: int = 0):
        super().__init__(name, pos, yaw, recipe_id)
        self.item_id = BuildingItem.QuantumChemicalPlant
        self.model_index = BuildingModel.QuantumChemicalPlant


from ..buildings import TeslaTower, ArcSmelter, AssemblingMachineMKI, ConveyorBeltMKI
from ..enums import Item
from ..utils import Yaw, Vector
from .factory_block import FactoryBlock
from .recipes import Recipe

import math

"""
A factory block consist of:
 - A single factory of any type
 - One to three input belts
 - One to three input sorters
 - One to three output belts
 - One to three output sorters
"""

class FactoryLine:
    
    def __init__(self, pos, belt_routing, recipe, factory_count):
        
        self.height = 6
        self._factory_type = FactoryLine.select_factory(recipe)
        self.block_width = int(self._get_factory_width())
        
        # Generate factory_blocks
        self.factory_blocks = []
        for i in range(factory_count):
            temp_pos = pos + Vector(x = i * self.block_width)
            factory_block = FactoryBlock(temp_pos, belt_routing, self._factory_type, self.block_width, recipe)
            self.factory_blocks.append(factory_block)
        
        # Connect factory_blocks
        for i in range(len(self.factory_blocks) - 1):
            self.factory_blocks[i].connect_to_factory_block(self.factory_blocks[i + 1])
        
        #self.input_belts = [factory_blocks[0].input_belts[i][0] for i in range(input_count)]
        #self.output_belts = [factory_blocks[0].output_belts[i][-1] for i in range(output_count)]
    
    def _get_factory_width(self):
        return self._factory_type.get_size().x
    
    def select_factory(recipe):
        if recipe["tool"] == "Smelting Facility":
            return ArcSmelter
        elif recipe["tool"] == "Assembling Machine":
            return AssemblingMachineMKI
        else:
            assert False, f"Unknown tool: {recipe['tool']}, Recipe: {recipe['name']}, ID: {recipe['recipe_id']}"


from .building import Building, Factory3x3
from ..utils import Vector, Yaw
from ..enums import BuildingItem, BuildingModel, Item

import enum

class DepotMKI(Factory3x3):
    def __init__(self, name, pos: Vector, blocked_slots: int = 0):
        super().__init__(name)
        self.pos = pos
        self.pos2 = pos
        self.yaw = Yaw.North
        self.yaw2 = Yaw.North
        self.item_id = BuildingItem.DepotMKI
        self.model_index = BuildingModel.DepotMKI
        self.output_object_index = -1
        self.input_object_index = -1
        self.parameter_count = 110
        self.parameters = [blocked_slots] + [0 for i in range(109)]
        
    def get_size():
        return Vector(4.0, 4.0)
    
    def get_offset():
        return Vector(2.0, 2.0)

    def create_logistic_distributor(self, name, filter_id, chargin_power, icarus_mode, distributor_mode):
        self.output_to_slot = 14
        self.input_from_slot = 15
        self.output_to_slot = 15
        self.input_from_slot = 14
        return LogisticDistributor(
            name = name,
            pos = self.pos + Vector(z = 1.875),
            input_object_index = self.index,
            filter_id = filter_id,
            charging_power = chargin_power,
            icarus_mode = icarus_mode,
            distributor_mode = distributor_mode
        )

class DepotMKII:
    pass
      
class LogisticDistributor(Building):
    
    """
    0 < charging_power < 9 [MW]
    """
    
    class IcarusMode(enum.IntEnum):
        DisableIcarus = 0
        CollectFromIcarus = 1
        CollectFromAndProvideToIcarus = 2
        ProvideToIcarus = 3
    
    class DistributorMode(enum.IntEnum):
        DisableDistributors = 0
        ProvideToDistributors = 1
        RequestFromDistributors = 2

    def __init__(self, name, pos: Vector, input_object_index, charging_power: int, filter_id: Item, icarus_mode: IcarusMode, distributor_mode: DistributorMode):
        super().__init__(name)
        self.pos = pos
        self.pos2 = pos
        self.yaw = Yaw.North
        self.yaw2 = Yaw.North
        self.item_id = BuildingItem.LogisticDistributor
        self.model_index = BuildingModel.LogisticDistributor
        self.input_object_index = input_object_index
        self.filter_id = filter_id
        self.output_to_slot = 0
        self.input_from_slot = 13
        self.input_to_slot = 0
        self.output_from_slot = 0
        self.parameter_count = 128
        self.parameters = [icarus_mode, distributor_mode, 150000] + [0 for i in range(125)]
        


from .packet import Packet
import random

class BlueprintStringHeader:
    SIZE = 10

    def __init__(self):
        self.fixed0_1 = 0
        self.layout = 10
        self.icon0 = 0
        self.icon1 = 0
        self.icon2 = 0
        self.icon3 = 0
        self.icon4 = 0
        self.fixed0_2 = 0
        self.timestamp = 0
        self.game_version = "0.10.29.22015"
        self.short_desc = "New%20Blueprint"

    def parse(self, string):
        component = string[BlueprintStringHeader.SIZE:].split(",")
        self.fixed0_1 = component[0]
        self.layout = component[1]
        self.icon0 = component[2]
        self.icon1 = component[3]
        self.icon2 = component[4]
        self.icon3 = component[5]
        self.icon4 = component[6]
        self.fixed0_2 = component[7]
        self.timestamp = component[8]
        self.game_version = component[9]
        self.short_desc = component[10]

    def serialize(self):
        return f"BLUEPRINT:{str(self.fixed0_1)},{str(self.layout)},{str(self.icon0)},{str(self.icon1)},{str(self.icon2)},{str(self.icon3)},{str(self.icon4)},{str(self.fixed0_2)},{str(self.timestamp)},{str(self.game_version)},{str(self.short_desc)},"

    def __str__(self):
        return f"""
Blueprint string header:
========================
Binary data: {self.serialize()}
========================
Fixed 1: {self.fixed0_1}
Layout: {self.layout}
Icon0: {self.icon0}
Icon1: {self.icon1}
Icon2: {self.icon2}
Icon3: {self.icon3}
Icon4: {self.icon4}
Fixed 2: {self.fixed0_2}
Timestamp: {self.timestamp}
Game version: {self.game_version}
Description: {self.short_desc}
"""


import enum

class SmeltingRecipe(enum.IntEnum):
    IronIngot = 1
    Magent = 2
    CopperIngot = 3
    StoneBrick = 4
    EnergeticGraphite = 17
    SiliconOre = 34
    CrystalSilicon = 37
    Glass = 57
    HighPuritySilicon = 59
    Diamond = 60
    DiamondAdvanced = 61
    Steel = 63
    TitaniumIngot = 65
    TitaniumAlloy = 66

class AssemblingRecipe(enum.IntEnum):
	Gear = 5
	MagneticCoil = 6
	WindTurbine = 7
	TeslaTower = 8
	MatrixLab = 10
	Prism = 11
	PlasmaExciter = 12
	WirelessPowerTower = 13
	OilExtractor = 14
	OilRefinary = 15
	HydrogenFuelRod = 19
	Thruster = 20
	ReinforcedThruster = 21
	ChemicalPlant = 22
	TitaniumCrystal = 26
	CashmirCrystal = 28
	CashmirCrystalAdvanced = 29
	TitaniumGlass = 30
	ParticleBroadband = 36
	PlaneFilter = 38
	MiniatureParticleCollider = 39
	DeuteronFuelRod = 41
	AnnihilationConstraintSphere = 42
	AntimatterFuelRod = 44
	AssemblingMachineMKI = 45
	AssemblingMachineMKII = 46
	AssemblingMachineMKIII = 47
	MiningMachine = 48
	WaterPump = 49
	CircuitBoard = 50
	Processor = 51
	QuantumChip = 52
	MicrocrystallineComponent = 53
	OrganicCrystal = 54
	ArcSmelter = 56
	CrystalSiliconAdvanced = 62
	ThermalPowerPlant = 64
	SolarPanel = 67
	PhotonCombiner = 68
	PhotonCombinerAdvanced = 69
	SolarSail = 70
	EMRailEjector = 71
	RayReceiver = 72
	SateliteSubstation = 73
	Accumulator = 76
	EnergyExchanger = 77
	SpaceWraper = 78
	SpaceWraperAdvanced = 79
	FrameMaterial = 80
	DysonSphereComponent = 81
	VerticalLaunchingSilo = 82
	DysonSphereRocket = 83
	ConveyorBeltMKI = 84
	SorterMKI = 85
	DepotMKI = 86
	Splitter = 87
	SorterMKII = 88
	ConveyorBeltMKII = 89
	DepotMKII = 91
	ConveyorBeltMKIII = 92
	PlanetaryLogisticsStation = 93
	LogisticDrone = 94
	InterstellarLogisticsStation = 95
	LogisticsVessel = 96
	ElectricMotor = 97
	ParticleContainer = 99
	ParticleContainerAdvanced = 100
	GravitationLens = 101
	SuperMagneticRing = 103
	Engine = 105
	ProlifiratorMKI = 106
	ProlifiratorMKII = 107
	ProlifiratorMKIII = 108
	SprayCoater = 109
	Fractionator = 110
	ObitalCollector = 111
	MiniFusionPowerPlant = 113
	StorageTank = 114
	PlaneSmelter = 116
	ItemMonitor = 117
	GeothermalPowerPlant = 118
	AdvancedMiningMachine = 119
	AutomaticPiler = 120
	LogisticDistributor = 122
	LogisticBot = 123
	QuantumChemicalPlant = 124
	GaussTurrent = 125
	LaserTurret = 126
	ImplosionCanon = 127
	PlasmaTurret = 128
	MissleTurret = 129
	JammerTower = 130
	SignalTower = 131
	PlanetaryShieldGenerator = 132
	CombustibleUnit = 133
	MagnumAmmoBox = 136
	TitaniumAmmoBox = 137
	SuperAlloyAmmoBox = 138
	ShellSet = 139
	HighExplosiveShellSet = 140
	CrystalShellSet = 141
	PlasmaCapsule = 142
	MissleSet = 144
	SuperSonicMissleSet = 145
	Prototype = 147
	PrecisionDrone = 148
	AttackDrone = 149
	Covette = 150
	Destroyer = 151
	BattleFieldAnalysisBase = 152
	SRPlasmaTurret = 157
	JamingCapsule = 158
	StackSorter = 160

class BuildingItem(enum.IntEnum):
    ConveyorBeltMKI = 2001
    ConveyorBeltMKII = 2002
    ConveyorBeltMKIII = 2003
    
    SorterMKI = 2011
    SorterMKII = 2012
    SorterMKIII = 2013
    PileSorter = 2014
    
    Splitter = 2020 # TODO implement
    
    TrafficMonitor = 2030 # TODO implement
    
    DepotMKI = 2101 # TODO implement
    DepotMKII = 2102 # TODO implement
    
    StorageTank = 2106 # TODO implement
    LogisticDistributor = 2107
    
    TeslaTower = 2201
    WirelessPowerTower = 2202
    WindTurbine = 2203 # TODO implement
    ThermalPowerPlant = 2204 # TODO implement
    SolarPanel = 2205 # TODO implement
    Accumulator = 2206 # TODO implement
    
    MiniFusionPowerPlant = 2211 # TODO implement
    SateliteSubstation = 2212
    
    ArcSmelter = 2302
    AssemblingMachineMKI = 2303
    AssemblingMachineMKII = 2304
    AssemblingMachineMKIII = 2305
    
    OilRefinary = 2308 # TODO implement
    ChemicalPlant = 2309 # TODO implement
    MiniatureParticleCollider = 2310 # TODO implement
    
    SprayCoater = 2313 # TODO implement
    Fractionator = 2314 # TODO implement
    
    QuantumChemicalPlant = 2317 # TODO implement

class BuildingModel(enum.IntEnum):
    ConveyorBeltMKI = 35
    ConveyorBeltMKII = 36
    ConveyorBeltMKIII = 37
    SplitterCross = 38
    SplitterTwoLayerStraight = 39
    SplitterTwoLayerCross = 40
    SorterMKI = 41
    SorterMKII = 42
    SorterMKIII = 43
    TeslaTower = 44
    
    Accumulator = 46
    
    DepotMKI = 51
    DepotMKII = 52
    WindTurbine = 53
    ThermalPowerPlant = 54
    SolarPanel = 55
    
    ArcSmelter = 62
    OilRefinary = 63
    ChemicalPlant = 64
    
    SateliteSubstation = 68
    MiniatureParticleCollider = 69
    
    WirelessPowerTower = 71
    
    AssemblingMachineMKI = 65
    AssemblingMachineMKII = 66
    AssemblingMachineMKIII = 67
    
    MiniFusionPowerPlant = 118
    Fractionator = 119
    SprayCoater = 120
    StorageTank = 121
    
    TrafficMonitor = 208
    LogisticDistributor = 371
    
    QuantumChemicalPlant = 376
    
    PileSorter = 483

class Item(enum.IntEnum):
	Lava = -1
	Water = 1000
	IronOre = 1001
	CopperOre = 1002
	SiliconOre = 1003
	TitaniumOre = 1004
	Stone = 1005
	Coal = 1006
	CrudeOil = 1007
	FireIce = 1011
	KimberliteOre = 1012
	FractalSilicon = 1013
	OpticalGratingCrystal = 1014
	SpiniformStalagmiteCrystal = 1015
	UnipolarMagnet = 1016
	Log = 1030
	PlantFuel = 1031
	IronIngot = 1101
	Magnet = 1102
	Steel = 1103
	CopperIngot = 1104
	HighPuritySilicon = 1105
	TitaniumIngot = 1106
	TitaniumAlloy = 1107
	StoneBrick = 1108
	EnergeticGraphite = 1109
	Glass = 1110
	Prism = 1111
	Diamond = 1112
	CrystalSilicon = 1113
	RefinedOil = 1114
	Plastic = 1115
	SulfuricAcid = 1116
	OrganicCrystal = 1117
	TitaniumCrystal = 1118
	TitaniumGlass = 1119
	Hydrogen = 1120
	Deuterium = 1121
	Antimatter = 1122
	Graphene = 1123
	CarbonNanotube = 1124
	FrameMaterial = 1125
	CasimirCrystal = 1126
	StrangeMatter = 1127
	Foundation = 1131
	AccelerantMkI = 1141
	AccelerantMkII = 1142
	AccelerantMkIII = 1143
	Gear = 1201
	MagneticCoil = 1202
	ElectricMotor = 1203
	ElectromagneticTurbine = 1204
	SuperMagneticRing = 1205
	ParticleContainer = 1206
	CriticalPhoton = 1208
	GravitationLens = 1209
	SpaceWarper = 1210
	CircuitBoard = 1301
	MicrocrystallineComponent = 1302
	Processor = 1303
	PlaneFilter = 1304
	QuantumChip = 1305
	PlasmaExciter = 1401
	ParticleBroadband = 1402
	AnnihilationConstraintSphere = 1403
	PhotonCombiner = 1404
	Thruster = 1405
	ReinforcedThruster = 1406
	SolarSail = 1501
	DysonSphereComponent = 1502
	SmallCarrierRocket = 1503
	HydrogenFuelRod = 1801
	DeuteronFuelRod = 1802
	AntimatterFuelRod = 1803
	ConveyorBeltMKI = 2001
	ConveyorBeltMKII = 2002
	ConveyorBeltMKIII = 2003
	SorterMKI = 2011
	SorterMKII = 2012
	SorterMKIII = 2013
	Splitter = 2020
	DepotMKI = 2101
	DepotMKII = 2102
	PlanetaryLogisticsStation = 2103
	InterstellarLogisticsStation = 2104
	OrbitalCollector = 2105
	StorageTank = 2106
	LogisticDistributor = 2107
	TeslaTower = 2201
	WirelessPowerTower = 2202
	WindTurbine = 2203
	ThermalPowerStation = 2204
	SolarPanel = 2205
	Accumulator = 2206
	AccumulatorFull = 2207
	RayReceiver = 2208
	EnergyExchanger = 2209
	ArtificialStar = 2210
	MiniFusionPowerStation = 2211
	SatelliteSubstation = 2212
	MiningMachine = 2301
	ArcSmelter = 2302
	AssemblingMachineMKI = 2303
	AssemblingMachineMKII = 2304
	AssemblingMachineMKIII = 2305
	WaterPump = 2306
	OilExtractor = 2307
	OilRefinery = 2308
	ChemicalPlant = 2309
	MiniatureParticleCollider = 2310
	EMRailEjector = 2311
	VerticalLaunchingSilo = 2312
	SprayCoater = 2313
	Fractionator = 2314
	PlaneSmelter = 2315
	MatrixLab = 2901
	LogisticsDrone = 5001
	LogisticsVessel = 5002
	ElectromagneticMatrix = 6001
	EnergyMatrix = 6002
	StructureMatrix = 6003
	InformationMatrix = 6004
	GravityMatrix = 6005
	UniverseMatrix = 6006
 
 


from .recipes import Recipe

class ItemFlow:

    def __init__(self, name, count_pr_sec):
        self.name = name
        self.count_pr_sec = count_pr_sec

    def select_recipe(self, item_name):
        return Recipe.select(item_name)

    def get_needed_ingredients(self, prolifirator = None):
        
        if prolifirator == "Mk.I":
            scale = 1.0/1.125
        elif prolifirator == "Mk.II":
            scale = 1.0/1.20
        elif prolifirator == "Mk.III":
            scale = 1.0/1.25
        else:
            scale = 1

        if not 'recipe' in vars(self):
            print(self.name)
            self.recipe = self.select_recipe(self.name)
            
        if self.recipe == None: # Assumes that this is a raw material
            return []

        ingredients = []

        for input_item, input_item_count in self.recipe["input_items"].items():
            ingredients.append(ItemFlow(input_item, input_item_count * self.count_pr_sec * scale / self.recipe["output_items"][self.name]))

        return ingredients
    
    def get_products(self):
        
        if not 'recipe' in vars(self):
            self.recipe = self.select_recipe(self.name)
        
        if self.recipe == None: # Assumes that this is a raw material
            return []

        products = []
        for output_item, output_item_count in self.recipe["output_items"].items():
            products.append(ItemFlow(output_item, output_item_count * self.count_pr_sec / self.recipe["output_items"][self.name]))

        return products


import yaml

import pkgutil

class Recipe:

    recipes = None

    def __init__(self, name, input_items, output_items, time, tool, recipe_id):
        self.name = name
        self.input_items = input_items
        self.output_items = output_items
        self.time = time
        self.tool = tool
        self.recipe_id = recipe_id

    def __str__(self):
        return f'Name: {self.name} - Input item: {self.input_items} - Output items: {self.output_items} - Time: {self.time}s - Tool: {self.tool} - Recipe id: {self.recipe_id}'

    def load_from_yaml(filename):
        data = pkgutil.get_data(__package__, filename)
        if data is not None:
            return yaml.safe_load(data)
        else:
            raise FileNotFoundError(f'{filename} not found in package')

    def select(item_name):
        for name in Recipe.recipes.keys():
            if item_name in name:
                print(name)
        if (item_name + "Advanced") in Recipe.recipes.keys():
            print(1)
        return Recipe.recipes[item_name]

Recipe.recipes = Recipe.load_from_yaml("data/recipes.yaml")

if __name__ == "__main__":
    filename = "data/recipes.yaml"
    recipes = load_from_yaml(filename)
    print(recipes)